

from matplotlib import pyplot as plt
import pandas, numpy
from .plotting import plot_as_svg_xhtml


def pseudo_bar_data(x_bins, y, gap=0):
	"""
	Parameters
	----------
	x_bins : array-like, shape=(N+1,)
		The bin boundaries
	y : array-like, shape=(N,)
		The bar heights

	Returns
	-------
	x, y
	"""
	# midpoints = (x_bins[1:] + x_bins[:-1]) / 2
	# widths = x_bins[1:] - x_bins[:-1]
	if gap:
		x_doubled = numpy.zeros(((x_bins.shape[0] - 1) * 4), dtype=numpy.float)
		x_doubled[::4] = x_bins[:-1]
		x_doubled[1::4] = x_bins[:-1]
		x_doubled[2::4] = x_bins[1:] - gap
		x_doubled[3::4] = x_bins[1:] - gap
		y_doubled = numpy.zeros(((y.shape[0]) * 4), dtype=y.dtype)
		y_doubled[1::4] = y
		y_doubled[2::4] = y
	else:
		x_doubled = numpy.zeros((x_bins.shape[0] - 1) * 2, dtype=x_bins.dtype)
		x_doubled[::2] = x_bins[:-1]
		x_doubled[1::2] = x_bins[1:]
		y_doubled = numpy.zeros((y.shape[0]) * 2, dtype=y.dtype)
		y_doubled[::2] = y
		y_doubled[1::2] = y
	return x_doubled, y_doubled



def distribution_figure(
		x,
		probability,
		choices=None,
		availability=None,
		xlabel=None,
		ylabel='Relative Frequency',
		style='hist',
		bins=None,
		pct_bins=20,
		range=None,
		prob_label="Modeled",
		obs_label="Observed",
		bw_method=None,
		discrete=None,
		ax=None,
		format='ax',
		accumulator=False,
		**kwargs,
):
	"""
	Generate a figure of observed and modeled choices over a range of variable values.

	Parameters
	----------
	x : array-like
		An array giving values for some variable.
	probability : array-like
		The pre-calculated probability array for all cases in this analysis.
		Must be the same shape as `x`.
	choices : array-like, optional
		The observed choices array for all cases in this analysis. If provided,
		must be the same shape as `x`.
	availability : array-like, optional
		The availability array for all cases in this analysis. If provided,
		must be the same shape as `x`.
	xlabel : str, optional
		A label to use for the x-axis of the resulting figure.  If not given,
		the value of `x.name` is used if it is defined.  Set to `False` to omit the
		x-axis label even if `x.name` is defined.
	ylabel : str, default "Relative Frequency"
		A label to use for the y-axis of the resulting figure.
	style : {'hist', 'kde'}
		The style of figure to produce, either a histogram or a kernel density plot.
	bins : int, default 25
		The number of bins to use, only applicable to histogram style.
	range : 2-tuple, optional
		A range to truncate the figure.
	prob_label : str, default "Modeled"
		A label to put in the legend for the modeled probabilities
	obs_label : str, default "Observed"
		A label to put in the legend for the observed choices
	subselector : str or array-like, optional
		A filter to apply to cases. If given as a string, this is loaded from the
		model's `dataservice` as an `idco` variable.
	ax : matplotlib.Axes, optional
		If given, the figure will be drawn on these axes and they will be returned,
		otherwise new blank axes are used to draw the figure.
	format : {'ax', 'figure', 'svg'}, default 'figure'
		How to return the result if it is a figure. The default is to return
		the raw matplotlib Axes instance.  Change this to `svg` to get a SVG
		rendering as an xmle.Elem.
	accumulator : bool, default False
		Add an net cumulative trend on the bottom.

	Returns
	-------
	Elem or Axes
		Returns `ax` if given as an argument, otherwise returns a rendering as an Elem
	"""

	if xlabel is None:
		try:
			xlabel = x.name
		except AttributeError:
			pass

	discrete_values = None
	if discrete:
		discrete_values = numpy.unique(x)
	elif discrete is None:
		from .histograms import seems_like_discrete_data
		discrete, discrete_values = seems_like_discrete_data(numpy.asarray(x).reshape(-1), return_uniques=True)

	x_discrete_labels = None if discrete_values is None else [str(i) for i in discrete_values]

	if bins is None:
		if x_discrete_labels is not None:
			# Discrete bins using defined labels
			bins = numpy.arange(len(x_discrete_labels)+1)
		if isinstance(x.dtype, pandas.CategoricalDtype):
			# Discrete bins using implied labels
			discrete_values = numpy.arange(len(x_discrete_labels))
			bins = numpy.arange(len(x_discrete_labels)+1)
			x = x.cat.codes
		else:
			x_ = x
			if availability is not None:
				x_ = x[availability != 0]
			low_pctile = 0
			high_pctile = 100
			if range:
				import scipy.stats
				if range[0] is not None:
					low_pctile = scipy.stats.percentileofscore(x_, range[0])
				if range[1] is not None:
					high_pctile = scipy.stats.percentileofscore(x_, range[1])
			if isinstance(pct_bins, int):
				bins = numpy.percentile(x_, numpy.linspace(low_pctile, high_pctile, pct_bins + 1))
			else:
				bins = numpy.percentile(x_, pct_bins)
	elif isinstance(bins, int) and availability is not None:
		# Equal width bin generation using only available alternatives
		x_ = x[availability != 0]
		if range:
			range_low, range_high = range
			if range_low is None:
				range_low = x_.min()
			if range_high is None:
				range_high = x_.max()
		else:
			range_low = x_.min()
			range_high = x_.max()
		bins = numpy.linspace(range_low, range_high, bins + 1)

	model_result = probability
	model_choice = choices

	if style == 'kde':
		import scipy.stats
		kernel_result = scipy.stats.gaussian_kde(x, bw_method=bw_method, weights=model_result.reshape(-1))
		common_bw = kernel_result.covariance_factor()
		if model_choice is not None:
			kernel_choice = scipy.stats.gaussian_kde(x, bw_method=common_bw, weights=model_choice.reshape(-1))
		else:
			kernel_choice = None

		if not range:
			x_ = x
			if availability is not None:
				x_ = x[availability != 0]
			range = (x_.min(), x_.max())

		x_points = numpy.linspace(*range, 250)
		y_points_1 = kernel_result(x_points)
		y_points_2 = kernel_choice(x_points)

	else:
		shift = 0.4 if discrete else 0
		gap = 0.2 if discrete else 0
		if range:
			range_low, range_high = range
			if range_low is None:
				range_low = x.min()
			if range_high is None:
				range_high = x.max()
			range = (range_low, range_high)

		y1, x1 = numpy.histogram(
			x,
			weights=model_result.reshape(-1),
			bins=bins,
			range=range,
			density=True,
		)
		x_points, y_points_1 = pseudo_bar_data(x1 - shift, y1, gap=gap)

		if model_choice is not None:
			y2, x2 = numpy.histogram(
				x,
				weights=model_choice.reshape(-1),
				bins=x1,
				density=True,
			)
			x_points, y_points_2 = pseudo_bar_data(x1 - shift, y2, gap=gap)

	if xlabel is False:
		xlabel = None

	if accumulator and model_choice is not None:
		fig, (ax, ax2) = plt.subplots(2, 1, gridspec_kw={'height_ratios': [3, 1], 'hspace':0}, sharex='col')
		cum_sum = (y1 - y2).cumsum()
		x1plus = (x1[:-1] + x1[1:])/2
		ax2.plot(x1plus, cum_sum, color='k', lw=0.5)
		ax2.fill_between(x1plus, numpy.fmin(cum_sum, 0), 0, facecolor='#ffa200', label=f"{prob_label} Ahead")
		ax2.fill_between(x1plus, numpy.fmax(cum_sum, 0), 0, facecolor='#1f77b4', label=f"{obs_label} Ahead")
		ax2.set_yticks([])
		ax2.set_ylabel("Net Cum.")
		ax2.legend()

	elif ax is None:
		fig, ax = plt.subplots()
	else:
		fig = None

	ax.bins = bins
	ax.plot(x_points, y_points_1, label=prob_label, lw=1.5)
	if model_choice is not None:
		ax.fill_between(x_points, y_points_2, label=obs_label, step=None, facecolor='#ffbe4d', edgecolor='#ffa200', lw=1.5)
	ax.legend()
	if not discrete:
		ax.set_xlim(x_points[0], x_points[-1])
	if x_discrete_labels is not None:
		ax.set_xticks(numpy.arange(len(x_discrete_labels)))
		ax.set_xticklabels(x_discrete_labels)
	if accumulator and model_choice is not None:
		ax2.set_xlabel(xlabel)
	else:
		ax.set_xlabel(xlabel)
	ax.set_yticks([])
	ax.set_ylabel(ylabel)
	if fig is None:
		return ax
	fig.tight_layout(pad=0.5)
	if format == 'svg':
		result = plot_as_svg_xhtml(fig, **kwargs)
		fig.clf()
		plt.close(fig)
	elif format == 'png':
		from .png import make_png
		result = make_png(fig, **kwargs)
		fig.clf()
		plt.close(fig)
	else:
		result = ax
	return result





def distribution_on_idca_variable(
		model,
		x,
		xlabel=None,
		ylabel='Relative Frequency',
		style='hist',
		bins=None,
		pct_bins=20,
		range=None,
		xlim=None,
		prob_label="Modeled",
		obs_label="Observed",
		subselector=None,
		probability=None,
		bw_method=None,
		discrete=None,
		ax=None,
		format='ax',
		**kwargs,
):
	"""
	Generate a figure of observed and modeled choices over a range of variable values.

	Parameters
	----------
	model : Model
		The discrete choice model to analyze.
	x : str or array-like
		The name of an `idca` variable, or an array giving its values.  If this name exactly
		matches that of an `idca` column in the model's loaded `dataframes`, then
		those values are used, otherwise the variable is loaded from the model's
		`dataservice`.
	xlabel : str, optional
		A label to use for the x-axis of the resulting figure.  If not given,
		the value of `x` is used if it is a string.  Set to `False` to omit the
		x-axis label.
	ylabel : str, default "Relative Frequency"
		A label to use for the y-axis of the resulting figure.
	style : {'hist', 'kde'}
		The style of figure to produce, either a histogram or a kernel density plot.
	bins : int, default 25
		The number of bins to use, only applicable to histogram style.
	range : 2-tuple, optional
		A range to truncate the figure. (alias `xlim`)
	prob_label : str, default "Modeled"
		A label to put in the legend for the modeled probabilities
	obs_label : str, default "Observed"
		A label to put in the legend for the observed choices
	subselector : str or array-like, optional
		A filter to apply to cases. If given as a string, this is loaded from the
		model's `dataservice` as an `idco` variable.
	probability : array-like, optional
		The pre-calculated probability array for all cases in this analysis.
		If not given, the probability array is calculated at the current parameter
		values.
	ax : matplotlib.Axes, optional
		If given, the figure will be drawn on these axes and they will be returned,
		otherwise new blank axes are used to draw the figure.
	format : {'ax', 'figure', 'svg'}, default 'figure'
		How to return the result if it is a figure. The default is to return
		the raw matplotlib Axes instance.  Change this to `svg` to get a SVG
		rendering as an xmle.Elem.

	Other Parameters
	----------------
	header : str, optional
		A header to attach to the figure.  The header is not generated using
		matplotlib, but instead is prepended to the xml output with a header tag before the
		rendered svg figure.


	Returns
	-------
	Elem or Axes
		Returns `ax` if given as an argument, otherwise returns a rendering as an Elem
	"""

	if model is None:
		return lambda mdl: distribution_on_idca_variable(
			mdl,
			xlabel=xlabel,
			ylabel=ylabel,
			style=style,
			bins=bins,
			pct_bins=pct_bins,
			range=range,
			xlim=xlim,
			prob_label=prob_label,
			obs_label=obs_label,
			subselector=subselector,
			probability=probability,
			bw_method=bw_method,
			discrete=discrete,
			ax=ax,
			format=format,
			**kwargs,
		)

	if xlim is not None and range is None:
		range = xlim

	if isinstance(x, str):
		x_label = x
		if model.dataframes and model.dataframes.data_ca_or_ce is not None and x in model.dataframes.data_ca_or_ce:
			x = model.dataframes.data_ca_or_ce[x].values.reshape(-1)
		else:
			x = model.dataservice.make_dataframes({'ca': [x]}, explicit=True).array_ca().reshape(-1)
	else:
		try:
			x_label = x.name
		except AttributeError:
			x_label = ''

	# if model.dataframes and model.dataframes.data_ca is not None and continuous_variable in model.dataframes.data_ca:
	# 	cv = model.dataframes.data_ca[continuous_variable].values.reshape(-1)
	# else:
	# 	cv = model.dataservice.make_dataframes({'ca': [continuous_variable]}, explicit=True).array_ca().reshape(-1)

	discrete_values = None
	if discrete:
		discrete_values = numpy.unique(x)
	elif discrete is None:
		from .histograms import seems_like_discrete_data
		discrete, discrete_values = seems_like_discrete_data(numpy.asarray(x).reshape(-1), return_uniques=True)

	x_discrete_labels = None if discrete_values is None else [str(i) for i in discrete_values]

	if bins is None:
		if x_discrete_labels is not None:
			# Discrete bins using defined labels
			bins = numpy.arange(len(x_discrete_labels)+1)
		if isinstance(x.dtype, pandas.CategoricalDtype):
			# Discrete bins using implied labels
			discrete_values = numpy.arange(len(x_discrete_labels))
			bins = numpy.arange(len(x_discrete_labels)+1)
			x = x.cat.codes
		else:
			x_ = x
			if model.dataframes.data_av is not None and model.dataframes.data_ca is not None:
				x_ = x[model.dataframes.data_av.values.reshape(-1) != 0]
			low_pctile = 0
			high_pctile = 100
			if range is not None:
				import scipy.stats
				if range[0] is not None:
					low_pctile = scipy.stats.percentileofscore(x_, range[0])
				if range[1] is not None:
					high_pctile = scipy.stats.percentileofscore(x_, range[1])
			if isinstance(pct_bins, int):
				bins = numpy.percentile(x_, numpy.linspace(low_pctile, high_pctile, pct_bins + 1))
			else:
				bins = numpy.percentile(x_, pct_bins)
	elif isinstance(bins, int) and model.dataframes.data_av is not None and model.dataframes.data_ca is not None:
		# Equal width bin generation using only available alternatives
		x_ = x[model.dataframes.data_av.values.reshape(-1) != 0]
		if range is not None:
			range_low, range_high = range
			if range_low is None:
				range_low = x_.min()
			if range_high is None:
				range_high = x_.max()
		else:
			range_low = x_.min()
			range_high = x_.max()
		bins = numpy.linspace(range_low, range_high, bins + 1)

	if probability is None:
		probability = model.probability()

	model_result = probability[:, :model.dataframes.n_alts]
	model_choice = model.dataframes.data_ch.values
	if model.dataframes.data_wt is not None:
		model_result = model_result.copy()
		model_result *= model.dataframes.data_wt.values.reshape(-1,1)
		model_choice = model_choice.copy()
		model_choice *= model.dataframes.data_wt.values.reshape(-1,1)

	if subselector is not None:
		if isinstance(subselector, str):
			subselector = model.dataservice.make_dataframes({'co': [subselector]}, explicit=True).array_co(dtype=bool).reshape(-1)
		x = numpy.asarray(x).reshape(*model_result.shape)[subselector].reshape(-1)
		model_result = model_result[subselector]
		model_choice = model_choice[subselector]

	if style == 'kde':
		import scipy.stats
		kernel_result = scipy.stats.gaussian_kde(x, bw_method=bw_method, weights=model_result.reshape(-1))
		common_bw = kernel_result.covariance_factor()
		kernel_choice = scipy.stats.gaussian_kde(x, bw_method=common_bw, weights=model_choice.reshape(-1))

		if range is None:
			x_ = x
			if model.dataframes.data_av is not None and model.dataframes.data_ca is not None:
				x_ = x[model.dataframes.data_av.values.reshape(-1) != 0]
			range = (x_.min(), x_.max())

		x_points = numpy.linspace(*range, 250)
		y_points_1 = kernel_result(x_points)
		y_points_2 = kernel_choice(x_points)

	else:
		if range is not None:
			range_low, range_high = range
			if range_low is None:
				range_low = x.min()
			if range_high is None:
				range_high = x.max()
			range = (range_low, range_high)

		y_points_1, x1 = numpy.histogram(
			x,
			weights=model_result.reshape(-1),
			bins=bins,
			range=range,
			density=True,
		)

		y_points_2, x2 = numpy.histogram(
			x,
			weights=model_choice.reshape(-1),
			bins=x1,
			density=True,
		)

		shift = 0.4 if discrete else 0
		gap = 0.2 if discrete else 0

		x_points, y_points_1 = pseudo_bar_data(x1 - shift, y_points_1, gap=gap)
		x_points, y_points_2 = pseudo_bar_data(x1 - shift, y_points_2, gap=gap)


	if xlabel is None:
		xlabel = x_label
	if xlabel is False:
		xlabel = None

	if ax is None:
		fig, ax = plt.subplots()
	else:
		fig = None

	ax.bins = bins
	ax.plot(x_points, y_points_1, label=prob_label, lw=1.5)
	ax.fill_between(x_points, y_points_2, label=obs_label, step=None, facecolor='#ffbe4d', edgecolor='#ffa200', lw=1.5)
	ax.legend()
	if not discrete:
		ax.set_xlim(x_points[0], x_points[-1])
	if x_discrete_labels is not None:
		ax.set_xticks(numpy.arange(len(x_discrete_labels)))
		ax.set_xticklabels(x_discrete_labels)
	ax.set_xlabel(xlabel)
	ax.set_yticks([])
	ax.set_ylabel(ylabel)
	if fig is None:
		return ax
	fig.tight_layout(pad=0.5)
	if format == 'svg':
		result = plot_as_svg_xhtml(fig, **kwargs)
		fig.clf()
		plt.close(fig)
	elif format == 'png':
		from .png import make_png
		result = make_png(fig, **kwargs)
		fig.clf()
		plt.close(fig)
	else:
		result = ax
	return result


from .. import Model

Model.distribution_on_idca_variable = distribution_on_idca_variable






def share_figure(
		x,
		probability,
		choices=None,
		weights=1,
		xlabel=None,
		bins=None,
		pct_bins=20,
		figsize=(12, 4),
		style='stacked',
		discrete=None,
		xlim=None,
		xscale=None,
		xmajorticks=None,
		xminorticks=None,
		include_nests=False,
		exclude_alts=None,
		format='figure',
		**kwargs,
):
	"""
	Generate a figure of variables over a range of variable values.

	Parameters
	----------
	x : array-like, 1-d
		An array giving values for some variable.
	probability : array-like, 2-d
		The pre-calculated probability array for all cases in this analysis.
		First dimension must be the same shape as `x`.  The second dimension
		represents the alternatives (or similar).
	choices : array-like, optional
		The observed choices array for all cases in this analysis. If provided,
		the first dimension must be the same shape as `x`.  The second dimension
		represents the alternatives (or similar).
	weights : array-like, 1-d, optional
		The case weights for all cases in this analysis. If provided,
		the shape must be the same shape as `x`.
	xlabel : str, optional
		A label to use for the x-axis of the resulting figure.  If not given,
		the value of `x.name` is used if it exists.  Set to `False` to omit the
		x-axis label.
	bins : int, optional
		The number of equal-sized bins to use.
	pct_bins : int or array-like, default 20
		The number of equal-mass bins to use.
	style : {'stacked', 'dataframe', 'many'}
		The type of output to generate.
	discrete : bool, default False
		Whether to treat the data values explicitly as discrete (vs continuous)
		data.  This will change the styling and automatic bin generation.  If
		there are very few unique values, the data will be assumed to be
		discrete anyhow.
	xlim : 2-tuple, optional
		Explicitly set the range of values shown on the x axis of generated
		figures.  This can truncate long tails.  The actual histogram bins
		are not changed.
	include_nests : bool, default False
		Whether to include nests in the figure.
	exclude_alts : Collection, optional
		Alternatives to exclude from the figure.
	filter : str, optional
		A filter that will be used to select only a subset of cases.
	format : {'figure','svg'}, default 'figure'
		How to return the result if it is a figure. The default is to return
		the raw matplotlib Figure instance, ot set to `svg` to get a SVG
		rendering as an xmle.Elem.

	Returns
	-------
	Figure, DataFrame, or Elem
	"""

	if style not in {'stacked', 'dataframe', 'many'}:
		raise ValueError("style must be in {'stacked', 'dataframe', 'many'}")

	if include_nests and style == 'stacked' and exclude_alts is None:
		import warnings
		warnings.warn("including nests in a stacked figure is likely to give "
					  "misleading results unless constituent alternatives are omitted")

	if exclude_alts is None:
		exclude_alts = set()

	if xlabel is None:
		try:
			xlabel = x.name
		except AttributeError:
			pass

	filter_ = slice(None)

	h_pr = {}
	h_ch = {}

	discrete_values = None
	if discrete:
		discrete_values = numpy.unique(x)
	elif discrete is None:
		from .histograms import seems_like_discrete_data
		discrete, discrete_values = seems_like_discrete_data(x, return_uniques=True)

	pr = numpy.asarray(probability)
	if choices is not None:
		ch = numpy.asarray(choices)
	else:
		ch = None
	wt = numpy.asarray(weights)

	x_discrete_labels = None if discrete_values is None else [str(i) for i in discrete_values]

	if bins is None:
		if isinstance(x.dtype, pandas.CategoricalDtype):
			discrete_values = numpy.arange(len(x_discrete_labels))
			bins = numpy.arange(len(x_discrete_labels)+1)
			x = x.cat.codes
		elif isinstance(pct_bins, int):
			bins = numpy.percentile(x, numpy.linspace(0, 100, pct_bins + 1))
		else:
			bins = numpy.percentile(x, pct_bins)

	try:
		columns = probability.columns
	except AttributeError:
		columns = None
	else:
		columns = dict(enumerate(columns))

	for i in range(pr.shape[1]):

		h_pr[i], _ = numpy.histogram(
			x,
			weights=pr[:, i] * wt,
			bins=bins,
		)
		if ch is not None:
			h_ch[i], _ = numpy.histogram(
				x,
				weights=ch[:, i] * wt,
				bins=bins,
			)

	h_pr = pandas.DataFrame(h_pr)
	h_pr.index = pandas.IntervalIndex.from_breaks(bins) # bins[:-1]
	h_pr.rename(columns=columns, inplace=True)
	_denominator, _ = numpy.histogram(
		x,
		weights=pr.sum(1) * wt,
		bins=bins,
	)
	h_pr_share = (h_pr / _denominator.reshape(-1, 1))
	if ch is not None:
		_denominator_ch, _ = numpy.histogram(
			x,
			weights=ch.sum(1) * wt,
			bins=bins,
		)
		h_ch = pandas.DataFrame(h_ch)
		h_ch.index = h_pr.index
		h_ch.rename(columns=columns, inplace=True)
		h_ch_share = (h_ch / _denominator_ch.reshape(-1, 1))
	else:
		h_ch_share = None

	if discrete:
		x_placement = numpy.arange(len(bins)-1)
		x_alignment = 'center'
		bin_widths = 0.8
	else:
		x_placement = bins[:-1]
		x_alignment = 'edge'
		bin_widths = bins[1:] - bins[:-1]

	if xlabel is False:
		xlabel = None

	if xlim is None:
		xlim = (bins[0], bins[-1])

	if style == 'dataframe':

		if ch is not None:
			result = pandas.concat({
				'Modeled Shares': h_pr_share,
				'Observed Shares': h_ch_share,
			}, axis=1, sort=False)
		else:
			result = pandas.concat({
				'Modeled Shares': h_pr_share,
			}, axis=1, sort=False)
		result['Count', '*'] = h_pr.sum(1)

		if xlabel:
			result.index.name = xlabel

	elif style == 'stacked':

		fig, (ax0, ax1) = plt.subplots(1, 2, figsize=figsize)

		bottom0 = 0
		bottom1 = 0

		for i in h_pr_share.columns:
			ax0.bar(
				x_placement,
				height=h_pr_share[i],
				bottom=bottom0,
				width=bin_widths,
				align=x_alignment,
				label=i,
			)
			bottom0 = h_pr_share[i].fillna(0).values + bottom0
			ax1.bar(
				x_placement,
				height=h_ch_share[i],
				bottom=bottom1,
				width=bin_widths,
				align=x_alignment,
			)
			bottom1 = h_ch_share[i].fillna(0).values + bottom1

		ax0.set_ylim(0, 1)
		if not discrete:
			ax0.set_xlim(*xlim)
			if xscale:
				ax0.set_xscale(xscale)
			if xmajorticks is not None:
				ax0.set_xticks(xmajorticks)
				ax0.set_xticklabels(xmajorticks)
			if xminorticks is not None:
				ax0.set_xticks(xminorticks, minor=True)
		if x_discrete_labels is not None:
			ax0.set_xticks(numpy.arange(len(x_discrete_labels)))
			ax0.set_xticklabels(x_discrete_labels)
		ax0.set_title('Modeled Shares')

		ax1.set_ylim(0, 1)
		if not discrete:
			ax1.set_xlim(*xlim)
			if xscale:
				ax1.set_xscale(xscale)
			if xmajorticks is not None:
				ax1.set_xticks(xmajorticks)
				ax1.set_xticklabels(xmajorticks)
			if xminorticks is not None:
				ax1.set_xticks(xminorticks, minor=True)
		if x_discrete_labels is not None:
			ax1.set_xticks(numpy.arange(len(x_discrete_labels)))
			ax1.set_xticklabels(x_discrete_labels)
		ax1.set_title('Observed Shares')
		if xlabel:
			ax0.set_xlabel(xlabel)
			ax1.set_xlabel(xlabel)

		fig.legend(
			loc='center right',
		)

		# fig.tight_layout(pad=0.5)
		if format == 'svg':
			result = plot_as_svg_xhtml(fig, **kwargs)
			fig.clf()
			plt.close(fig)
		elif format == 'png':
			from .png import make_png
			result = make_png(fig, **kwargs)
			fig.clf()
			plt.close(fig)
		else:
			result = fig

	else:

		fig, axes = plt.subplots(len(h_pr_share.columns), 1, figsize=figsize)

		shift = 0.4 if discrete else 0

		for n,i in enumerate(h_pr_share.columns):
			x_, y_ = pseudo_bar_data(bins-shift, h_pr_share[i], gap=0.2 if discrete else 0)
			axes[n].plot(x_, y_, label='Modeled' if n==0 else None, lw=1.5)

			x_ch_, y_ch_ = pseudo_bar_data(bins-shift, h_ch_share[i], gap=0.2 if discrete else 0)
			axes[n].fill_between(
				x_ch_, y_ch_, label='Observed' if n==0 else None, step=None,
				facecolor='#ffbe4d', edgecolor='#ffa200',
				lw=1.5,
			)
			if not discrete:
				axes[n].set_xlim(*xlim)
				if xscale:
					axes[n].set_xscale(xscale)
				if xmajorticks is not None:
					axes[n].set_xticks(xmajorticks)
					axes[n].set_xticklabels(xmajorticks)
				if xminorticks is not None:
					axes[n].set_xticks(xminorticks, minor=True)
			if x_discrete_labels is not None:
				axes[n].set_xticks(numpy.arange(len(x_discrete_labels)))
				axes[n].set_xticklabels(x_discrete_labels)
			axes[n].set_ylabel(i)

			# axes[n].legend(
			# 	# loc='center right',
			# )

		legnd = axes[0].legend(
			loc='lower center',
			ncol=2,
			borderaxespad=0,
			bbox_to_anchor=(0.5, 1.08)
		)

		if xlabel:
			axes[-1].set_xlabel(xlabel)
		#fig.tight_layout(pad=0.5)
		if format == 'svg':
			result = plot_as_svg_xhtml(fig, bbox_extra_artists=[legnd],  **kwargs)
			fig.clf()
			plt.close(fig)
		elif format == 'png':
			from .png import make_png
			result = make_png(fig, **kwargs)
			fig.clf()
			plt.close(fig)
		else:
			result = fig

	return result



def distribution_on_idco_variable(
		model,
		x,
		xlabel=None,
		bins=None,
		pct_bins=20,
		figsize=(12, 4),
		style='stacked',
		discrete=None,
		xlim=None,
		include_nests=False,
		exclude_alts=None,
		filter=None,
		format='figure',
		**kwargs,
):
	"""
	Generate a figure of variables over a range of variable values.

	Parameters
	----------
	model : Model
		The discrete choice model to analyze.
	x : str or array-like
		The name of an `idco` variable, or an array giving its values.  If this name exactly
		matches that of an `idco` column in the model's loaded `dataframes`, then
		those values are used, otherwise the variable is loaded from the model's
		`dataservice`.
	xlabel : str, optional
		A label to use for the x-axis of the resulting figure.  If not given,
		the value of `x` is used if it is a string.  Set to `False` to omit the
		x-axis label.
	bins : int, optional
		The number of equal-sized bins to use.
	pct_bins : int or array-like, default 20
		The number of equal-mass bins to use.
	style : {'stacked', 'dataframe', 'many'}
		The type of output to generate.
	discrete : bool, default False
		Whether to treat the data values explicitly as discrete (vs continuous)
		data.  This will change the styling and automatic bin generation.  If
		there are very few unique values, the data will be assumed to be
		discrete anyhow.
	xlim : 2-tuple, optional
		Explicitly set the range of values shown on the x axis of generated
		figures.  This can truncate long tails.  The actual histogram bins
		are not changed.
	include_nests : bool, default False
		Whether to include nests in the figure.
	exclude_alts : Collection, optional
		Alternatives to exclude from the figure.
	filter : str, optional
		A filter that will be used to select only a subset of cases.
	format : {'figure','svg'}, default 'figure'
		How to return the result if it is a figure. The default is to return
		the raw matplotlib Figure instance, ot set to `svg` to get a SVG
		rendering as an xmle.Elem.

	Returns
	-------
	Figure, DataFrame, or Elem
	"""

	if style not in {'stacked', 'dataframe', 'many'}:
		raise ValueError("style must be in {'stacked', 'dataframe', 'many'}")

	if include_nests and style == 'stacked' and exclude_alts is None:
		import warnings
		warnings.warn("including nests in a stacked figure is likely to give "
					  "misleading results unless constituent alternatives are omitted")

	if exclude_alts is None:
		exclude_alts = set()

	if isinstance(x, str):
		x_label = x
		if model.dataframes and model.dataframes.data_co is not None and x in model.dataframes.data_co:
			x = model.dataframes.data_co[x].values.reshape(-1)
		else:
			x = model.dataservice.make_dataframes({'co': [x]}, explicit=True).array_co().reshape(-1)
	else:
		try:
			x_label = x.name
		except AttributeError:
			x_label = ''

	if filter:
		_ds = model.dataservice if model.dataservice is not None else model.dataframes
		filter_ = _ds.make_dataframes(
			{'co': [filter]},
			explicit=True,
			float_dtype=bool,
		).array_co().reshape(-1)
		x = x[filter_]
	else:
		filter_ = slice(None)

	h_pr = {}
	h_ch = {}

	discrete_values = None
	if discrete:
		discrete_values = numpy.unique(x)
	elif discrete is None:
		from .histograms import seems_like_discrete_data
		discrete, discrete_values = seems_like_discrete_data(x, return_uniques=True)

	pr = model.probability(
		return_dataframe='names',
		include_nests=bool(include_nests),
	).loc[filter_,:]

	if include_nests:
		ch = model.dataframes.data_ch_cascade(model.graph).loc[filter_,:]
	else:
		ch = model.dataframes.data_ch.loc[filter_,:]


	if model.dataframes.data_wt is None:
		wt = 1
	else:
		wt = model.dataframes.data_wt.values.reshape(-1)[filter_]

	x_discrete_labels = None if discrete_values is None else [str(i) for i in discrete_values]

	if bins is None:
		if isinstance(x.dtype, pandas.CategoricalDtype):
			discrete_values = numpy.arange(len(x_discrete_labels))
			bins = numpy.arange(len(x_discrete_labels)+1)
			x = x.cat.codes
		elif isinstance(pct_bins, int):
			bins = numpy.percentile(x, numpy.linspace(0, 100, pct_bins + 1))
		else:
			bins = numpy.percentile(x, pct_bins)

	n_alts = model.graph.n_elementals()
	columns = {}

	for i in range(pr.shape[1]):
		columns[i] = pr.columns[i]
		if i < n_alts or include_nests is True or model.graph.standard_sort[i] in include_nests:
			if model.graph.standard_sort[i] == model.graph.root_id:
				continue
			if model.graph.standard_sort[i] in exclude_alts:
				continue
			h_pr[i], _ = numpy.histogram(
				x,
				weights=pr.iloc[:, i] * wt,
				bins=bins,
			)
			h_ch[i], _ = numpy.histogram(
				x,
				weights=ch.iloc[:, i] * wt,
				bins=bins,
			)

	h_pr = pandas.DataFrame(h_pr)
	h_pr.index = pandas.IntervalIndex.from_breaks(bins) # bins[:-1]
	h_pr.rename(columns=columns, inplace=True)
	_denominator, _ = numpy.histogram(
		x,
		weights=numpy.ones_like(pr.iloc[:, -1]) * wt,
		bins=bins,
	)
	h_pr_share = (h_pr / _denominator.reshape(-1, 1))
	h_ch = pandas.DataFrame(h_ch)
	h_ch.index = h_pr.index
	h_ch.rename(columns=columns, inplace=True)
	h_ch_share = (h_ch / _denominator.reshape(-1, 1))

	if discrete:
		x_placement = numpy.arange(len(bins)-1)
		x_alignment = 'center'
		bin_widths = 0.8
	else:
		x_placement = bins[:-1]
		x_alignment = 'edge'
		bin_widths = bins[1:] - bins[:-1]

	if xlabel is None:
		xlabel = x_label
	if xlabel is False:
		xlabel = None

	if xlim is None:
		xlim = (bins[0], bins[-1])

	if style == 'dataframe':

		result = pandas.concat({
			'Modeled Shares': h_pr_share,
			'Observed Shares': h_ch_share,
		}, axis=1, sort=False)
		result['Count', '*'] = h_pr.sum(1)

		if x_label:
			result.index.name = xlabel

	elif style == 'stacked':

		fig, (ax0, ax1) = plt.subplots(1, 2, figsize=figsize)

		bottom0 = 0
		bottom1 = 0

		for i in h_pr_share.columns:
			ax0.bar(
				x_placement,
				height=h_pr_share[i],
				bottom=bottom0,
				width=bin_widths,
				align=x_alignment,
				label=i,
			)
			bottom0 = h_pr_share[i].values + bottom0
			ax1.bar(
				x_placement,
				height=h_ch_share[i],
				bottom=bottom1,
				width=bin_widths,
				align=x_alignment,
			)
			bottom1 = h_ch_share[i].values + bottom1

		ax0.set_ylim(0, 1)
		if not discrete:
			ax0.set_xlim(*xlim)
		if x_discrete_labels is not None:
			ax0.set_xticks(numpy.arange(len(x_discrete_labels)))
			ax0.set_xticklabels(x_discrete_labels)
		ax0.set_title('Modeled Shares')

		ax1.set_ylim(0, 1)
		if not discrete:
			ax1.set_xlim(*xlim)
		if x_discrete_labels is not None:
			ax1.set_xticks(numpy.arange(len(x_discrete_labels)))
			ax1.set_xticklabels(x_discrete_labels)
		ax1.set_title('Observed Shares')
		if xlabel:
			ax0.set_xlabel(xlabel)
			ax1.set_xlabel(xlabel)

		fig.legend(
			loc='center right',
		)

		# fig.tight_layout(pad=0.5)
		if format == 'svg':
			result = plot_as_svg_xhtml(fig, **kwargs)
			fig.clf()
			plt.close(fig)
		elif format == 'png':
			from .png import make_png
			result = make_png(fig, **kwargs)
			fig.clf()
			plt.close(fig)
		else:
			result = fig

	else:

		fig, axes = plt.subplots(len(h_pr_share.columns), 1, figsize=figsize)

		shift = 0.4 if discrete else 0

		for n,i in enumerate(h_pr_share.columns):
			x_, y_ = pseudo_bar_data(bins-shift, h_pr_share[i], gap=0.2 if discrete else 0)
			axes[n].plot(x_, y_, label='Modeled' if n==0 else None, lw=1.5)

			x_ch_, y_ch_ = pseudo_bar_data(bins-shift, h_ch_share[i], gap=0.2 if discrete else 0)
			axes[n].fill_between(
				x_ch_, y_ch_, label='Observed' if n==0 else None, step=None,
				facecolor='#ffbe4d', edgecolor='#ffa200',
				lw=1.5,
			)
			if not discrete:
				axes[n].set_xlim(*xlim)
			if x_discrete_labels is not None:
				axes[n].set_xticks(numpy.arange(len(x_discrete_labels)))
				axes[n].set_xticklabels(x_discrete_labels)
			axes[n].set_ylabel(i)

			# axes[n].legend(
			# 	# loc='center right',
			# )

		legnd = axes[0].legend(
			loc='lower center',
			ncol=2,
			borderaxespad=0,
			bbox_to_anchor=(0.5, 1.08)
		)

		if xlabel:
			axes[-1].set_xlabel(xlabel)
		#fig.tight_layout(pad=0.5)
		if format == 'svg':
			result = plot_as_svg_xhtml(fig, bbox_extra_artists=[legnd],  **kwargs)
			fig.clf()
			plt.close(fig)
		elif format == 'png':
			from .png import make_png
			result = make_png(fig, **kwargs)
			fig.clf()
			plt.close(fig)
		else:
			result = fig

	return result



Model.distribution_on_idco_variable = distribution_on_idco_variable




import tables as tb
import os
from pathlib import Path
from collections.abc import MutableMapping
import cloudpickle
import pickle
from .h5util import get_or_create_group

class H5Vault(MutableMapping):

	def __init__(self, filename=None, mode='a', vaultnode=None, *,
	             h5f=None, inmemory=False, temp=False,
	             complevel=1, complib='zlib',
	             ):

		if isinstance(filename, H5Vault):
			# Copy / Re-Class contructor
			x = filename
			self._vaultnode = x._vaultnode
			self._h5f_own = False
			return

		if isinstance(filename, tb.group.Group) and vaultnode is None:
			# Called with just a group node, use it
			vaultnode = filename
			filename = None

		if isinstance(filename, (str,Path)):
			filename = os.fspath(filename)
			if vaultnode is None:
				vaultnode = "_VAULT_"

		if isinstance(vaultnode, tb.group.Group):
			# Use the existing info from this group node, ignore all other inputs
			self._vaultnode = vaultnode
			filename = vaultnode._v_file.filename
			mode = vaultnode._v_file.mode
			self._h5f_own = False

		elif isinstance(vaultnode, str):

			# apply expanduser to filename to allow for home-folder based filenames
			if isinstance(filename,str):
				filename = os.path.expanduser(filename)

			if filename is None:
				temp = True
				from ..util.temporaryfile import TemporaryFile
				self._TemporaryFile = TemporaryFile(suffix='.h5d')
				filename = self._TemporaryFile.name

			if h5f is not None:
				self._h5f_own = False
				# self._vaultnode = self._h5f.get_node(vaultnode)
				top = self._h5f.get_node('/')
				self._vaultnode = get_or_create_group(self._h5f, top, vaultnode, skip_on_readonly=True)
			else:
				kwd = {}
				if inmemory or temp:
					kwd['driver']="H5FD_CORE"
				if temp:
					kwd['driver_core_backing_store']=0
				if complevel is not None:
					kwd['filters']=tb.Filters(complib=complib, complevel=complevel)
				self._h5f_obj = tb.open_file(filename, mode, **kwd)
				self._h5f_own = True
				top = self._h5f_obj.get_node('/')
				self._vaultnode = get_or_create_group(self._h5f_obj, top, vaultnode, skip_on_readonly=True)
		else:
			raise ValueError('must give groupnode as `str` or `tables.group.Group`')

	@property
	def _h5f(self):
		return self._vaultnode._v_file

	def __contains__(self, name):
		if not isinstance(name, str):
			raise TypeError(f'vault keys must be str not {type(name)}')
		name = name.replace('.','_')
		if name in self._vaultnode._v_children:
			return True
		return False

	def __setitem__(self, name, value):
		if not isinstance(name, str):
			raise TypeError('H5Vault requires str keys')
		name = name.replace('.','_')
		if name not in self._vaultnode._v_children:
			vault_bin = self._h5f.create_vlarray(self._vaultnode, name, tb.ObjectAtom())
		else:
			vault_bin = self._vaultnode._v_children[name]
		vault_bin.append(cloudpickle.dumps(value))

	def _get_item_or_older(self, name, index=-1):
		name_ = name.replace('.','_')
		if name_ not in self._vaultnode._v_children:
			raise KeyError(name_+' not in vault')
		else:
			vault_bin = self._vaultnode._v_children[name_]
		return pickle.loads(vault_bin[index])

	def __getitem__(self, name):
		return self._get_item_or_older(name)

	def keys(self):
		return self._vaultnode._v_children.keys()

	def wipe(self, regex='.*'):
		import re
		names = [name for name in self.keys() if re.search(regex,name)]
		for name in names:
			self._vaultnode._v_file.remove_node(self._vaultnode, name)

	def __delitem__(self, key):
		self.wipe(key)

	def __iter__(self):
		for k in self.keys():
			yield k

	def __len__(self):
		return len(self._vaultnode._v_children.keys())






import pandas
import numpy
import operator

from sklearn.preprocessing import OneHotEncoder
from typing import Mapping

def one_hot_encode(vector, dtype='float32', categories=None, index=None):

	if isinstance(vector, pandas.Series):
		index = vector.index
		vector = vector.values

	encoder = OneHotEncoder(
		categories='auto' if categories is None else [categories,],
		sparse=False,
		dtype=dtype,
	).fit(vector.reshape(-1,1))

	return pandas.DataFrame(
		data = encoder.transform(vector.reshape(-1,1)),
		columns=encoder.categories_,
		index=index
	)

def periodize(
		values,
		mapping=None,
		default=None,
		right=True,
		left=True,
		**kwargs,
):
	"""
	Label sections of a continuous variable.

	This function contrasts with `pandas.cut` in that
	there can be multiple non-contiguous sections of the
	underlying continuous interval that obtain the same
	categorical value.

	Parameters
	----------
	values : array-like
		The values to label. If given as a pandas.Series,
		the returned values will also be a Series,
		with a categorical dtype.
	mapping : Collection or Mapping
		A mapping, or a collection of 2-tuples giving
		key-value pairs (not necessarily unique keys).
		The keys (or first values) will be the new values,
		and the values (or second values) are 2-tuples
		giving upper and lower bounds.
	default : any, default None
		Keys not inside any labeled interval will get
		this value instead.
	right : bool, default True
		Whether to include the upper bound[s] in the
		intervals for labeling.
	left : bool, default True
		Whether to include the lower bound[s] in the
		intervals for labeling.
	**kwargs :
		Are added to `mapping`.

	Returns
	-------
	array-like

	Example
	-------
	>>> import pandas
	>>> h = pandas.Series(range(1,24))
	>>> periodize(h, default='OP', AM=(6.5, 9), PM=(16, 19))
	0     OP
	1     OP
	2     OP
	3     OP
	4     OP
	5     OP
	6     AM
	7     AM
	8     AM
	9     OP
	10    OP
	11    OP
	12    OP
	13    OP
	14    OP
	15    PM
	16    PM
	17    PM
	18    PM
	19    OP
	20    OP
	21    OP
	22    OP
	dtype: category
	Categories (3, object): [AM, OP, PM]
	"""
	if mapping is None:
		mapping = []

	if isinstance(mapping, Mapping):
		mapping = list(mapping.items())

	mapping.extend(kwargs.items())

	if isinstance(values, pandas.Series):
		x = pandas.Series(index=values.index, data=default)
	else:
		x = numpy.full(values.shape, default)

	if right:
		rop = operator.le
	else:
		rop = operator.lt

	if left:
		lop = operator.ge
	else:
		lop = operator.gt

	for k,(lowerbound,upperbound) in mapping:
		if lowerbound is None:
			lowerbound = -numpy.inf
		if upperbound is None:
			upperbound = numpy.inf
		x[lop(values,lowerbound) & rop(values,upperbound)] = k

	if isinstance(x, pandas.Series):
		x = x.astype('category')

	return x




# The MIT License (MIT)
#
# Copyright (c) 2018 ESSS
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# https://github.com/ESSS/pytest-regressions/blob/master/src/pytest_regressions/dataframe_regression.py

import pytest
from pytest_regressions.common import perform_regression_check, import_error_message
import numpy as np

class DataFrameRegressionFixture:
	"""
	Pandas DataFrame Regression fixture implementation used on dataframe_regression fixture.
	"""

	DISPLAY_PRECISION = 17  # Decimal places
	DISPLAY_WIDTH = 1000  # Max. Chars on outputs
	DISPLAY_MAX_COLUMNS = 1000  # Max. Number of columns (see #3)

	def __init__(self, datadir, original_datadir, request):
		"""
		:type datadir: Path
		:type original_datadir: Path
		:type request: FixtureRequest
		"""
		self._tolerances_dict = {}
		self._default_tolerance = {}

		self.request = request
		self.datadir = datadir
		self.original_datadir = original_datadir
		self._force_regen = False

		self._pandas_display_options = (
			"display.precision",
			DataFrameRegressionFixture.DISPLAY_PRECISION,
			"display.width",
			DataFrameRegressionFixture.DISPLAY_WIDTH,
			"display.max_columns",
			DataFrameRegressionFixture.DISPLAY_MAX_COLUMNS,
		)

	def _check_data_types(self, key, obtained_column, expected_column):
		"""
		Check if data type of obtained and expected columns are the same. Fail if not.
		Helper method used in _check_fn method.
		"""
		try:
			import numpy as np
		except ModuleNotFoundError:
			raise ModuleNotFoundError(import_error_message("Numpy"))

		__tracebackhide__ = True
		obtained_data_type = obtained_column.values.dtype
		expected_data_type = expected_column.values.dtype
		if obtained_data_type != expected_data_type:
			# Check if both data types are comparable as numbers (float, int, short, bytes, etc...)
			if np.issubdtype(obtained_data_type, np.number) and np.issubdtype(
				expected_data_type, np.number
			):
				return

			# In case they are not, assume they are not comparable
			error_msg = (
				"Data type for data %s of obtained and expected are not the same.\n"
				"Obtained: %s\n"
				"Expected: %s\n" % (key, obtained_data_type, expected_data_type)
			)
			raise AssertionError(error_msg)

	def _check_data_shapes(self, obtained_column, expected_column):
		"""
		Check if obtained and expected columns have the same size.
		Helper method used in _check_fn method.
		"""
		__tracebackhide__ = True

		obtained_data_shape = obtained_column.values.shape
		expected_data_shape = expected_column.values.shape
		if obtained_data_shape != expected_data_shape:
			error_msg = (
				"Obtained and expected data shape are not the same.\n"
				"Obtained: %s\n"
				"Expected: %s\n" % (obtained_data_shape, expected_data_shape)
			)
			raise AssertionError(error_msg)

	def _check_fn(self, obtained_filename, expected_filename):
		"""
		Check if dict contents dumped to a file match the contents in expected file.
		:param str obtained_filename:
		:param str expected_filename:
		"""
		try:
			import numpy as np
		except ModuleNotFoundError:
			raise ModuleNotFoundError(import_error_message("Numpy"))
		try:
			import pandas as pd
		except ModuleNotFoundError:
			raise ModuleNotFoundError(import_error_message("Pandas"))

		__tracebackhide__ = True

		obtained_data = pd.read_csv(str(obtained_filename))
		expected_data = pd.read_csv(str(expected_filename))

		comparison_tables_dict = {}
		for k in obtained_data.keys():
			obtained_column = obtained_data[k]
			expected_column = expected_data.get(k)

			if expected_column is None:
				error_msg = f"Could not find key '{k}' in the expected results.\n"
				error_msg += "Keys in the obtained data table: ["
				for k in obtained_data.keys():
					error_msg += f"'{k}', "
				error_msg += "]\n"
				error_msg += "Keys in the expected data table: ["
				for k in expected_data.keys():
					error_msg += f"'{k}', "
				error_msg += "]\n"
				error_msg += "To update values, use --force-regen option.\n\n"
				raise AssertionError(error_msg)

			tolerance_args = self._tolerances_dict.get(k, self._default_tolerance)

			self._check_data_types(k, obtained_column, expected_column)
			self._check_data_shapes(obtained_column, expected_column)

			data_type = obtained_column.values.dtype
			if data_type in [float, np.float, np.float16, np.float32, np.float64]:
				not_close_mask = ~np.isclose(
					obtained_column.values,
					expected_column.values,
					equal_nan=True,
					**tolerance_args,
				)
			else:
				not_close_mask = obtained_column.values != expected_column.values

			if np.any(not_close_mask):
				diff_ids = np.where(not_close_mask)[0]
				diff_obtained_data = obtained_column[diff_ids]
				diff_expected_data = expected_column[diff_ids]
				if data_type == np.bool:
					diffs = np.logical_xor(obtained_column, expected_column)[diff_ids]
				else:
					diffs = np.abs(obtained_column - expected_column)[diff_ids]

				comparison_table = pd.concat(
					[diff_obtained_data, diff_expected_data, diffs], axis=1
				)
				comparison_table.columns = [f"obtained_{k}", f"expected_{k}", "diff"]
				comparison_tables_dict[k] = comparison_table

		if len(comparison_tables_dict) > 0:
			error_msg = "Values are not sufficiently close.\n"
			error_msg += "To update values, use --force-regen option.\n\n"
			for k, comparison_table in comparison_tables_dict.items():
				error_msg += f"{k}:\n{comparison_table}\n\n"
			raise AssertionError(error_msg)

	def _dump_fn(self, data_object, filename):
		"""
		Dump dict contents to the given filename
		:param pd.DataFrame data_object:
		:param str filename:
		"""
		data_object.to_csv(
			str(filename),
			float_format=f"%.{DataFrameRegressionFixture.DISPLAY_PRECISION}g",
		)

	def check(
		self,
		data_frame,
		basename=None,
		fullpath=None,
		tolerances=None,
		default_tolerance=None,
	):
		"""
		Checks the given pandas dataframe against a previously recorded version, or generate a new file.
		Example::
			data_frame = pandas.DataFrame.from_dict({
				'U_gas': U[0][positions],
				'U_liquid': U[1][positions],
				'gas_vol_frac [-]': vol_frac[0][positions],
				'liquid_vol_frac [-]': vol_frac[1][positions],
				'P': Pa_to_bar(P)[positions],
			})
			dataframe_regression.check(data_frame)
		:param pandas.DataFrame data_frame: pandas DataFrame containing data for regression check.
		:param str basename: basename of the file to test/record. If not given the name
			of the test is used.
		:param str fullpath: complete path to use as a reference file. This option
			will ignore embed_data completely, being useful if a reference file is located
			in the session data dir for example.
		:param dict tolerances: dict mapping keys from the data_dict to tolerance settings for the
			given data. Example::
				tolerances={'U': Tolerance(atol=1e-2)}
		:param dict default_tolerance: dict mapping the default tolerance for the current check
			call. Example::
				default_tolerance=dict(atol=1e-7, rtol=1e-18).
			If not provided, will use defaults from numpy's ``isclose`` function.
		``basename`` and ``fullpath`` are exclusive.
		"""
		try:
			import pandas as pd
		except ModuleNotFoundError:
			raise ModuleNotFoundError(import_error_message("Pandas"))

		import functools

		__tracebackhide__ = True

		if isinstance(data_frame, np.ndarray):
			data_frame = pd.DataFrame(data_frame)

		assert type(data_frame) is pd.DataFrame, (
			"Only pandas DataFrames are supported on on dataframe_regression fixture.\n"
			"Object with type '%s' was given." % (str(type(data_frame)),)
		)

		for column in data_frame.columns:
			array = data_frame[column]
			# Skip assertion if an array of strings
			if (array.dtype == "O") and (type(array[0]) is str):
				continue
			# Rejected: timedelta, datetime, objects, zero-terminated bytes, unicode strings and raw data
			assert array.dtype not in ["m", "M", "O", "S", "a", "U", "V"], (
				"Only numeric data is supported on dataframe_regression fixture.\n"
				"Array with type '%s' was given.\n" % (str(array.dtype),)
			)

		if tolerances is None:
			tolerances = {}
		self._tolerances_dict = tolerances

		if default_tolerance is None:
			default_tolerance = {}
		self._default_tolerance = default_tolerance

		dump_fn = functools.partial(self._dump_fn, data_frame)

		with pd.option_context(*self._pandas_display_options):
			perform_regression_check(
				datadir=self.datadir,
				original_datadir=self.original_datadir,
				request=self.request,
				check_fn=self._check_fn,
				dump_fn=dump_fn,
				extension=".csv.gz",
				basename=basename,
				fullpath=fullpath,
				force_regen=self._force_regen,
			)

@pytest.fixture
def dataframe_regression(datadir, original_datadir, request):
	"""
	Example usage:
	def testSomeData(num_regression):
		dataframe_regression.check(
			pandas.DataFrame.from_dict(
				{
					'U_gas': U[0],
					'U_liquid': U[1],
					'gas_vol_frac [-]': α[0],
					'liquid_vol_frac [-]': α[1],
					'P': Pa_to_bar(P),
				}
			),
			default_tolerance=dict(atol=1e-8, rtol=1e-8)
		)
	:type embed_data: _EmbedDataFixture
	:type request: FixtureRequest
	:rtype: DataRegressionFixture
	:return: Data regression fixture.
	"""
	return DataFrameRegressionFixture(datadir, original_datadir, request)



import larch, pytest
from larch import P,X,PX
from pytest import approx, raises
import numpy
import pandas

def test_nan_weight():
	m = larch.Model.Example(1)
	m.weight_co_var = 'hhowndum'
	m.load_data()
	m.dataframes.data_wt = m.dataframes.data_wt.div(m.dataframes.data_wt)

	m, diagnosis = m.doctor()
	assert type(m) is larch.Model
	assert 'nan_weight' in diagnosis
	assert diagnosis.nan_weight['n'].iloc[0] == 1882

	m, diagnosis = m.doctor(repair_nan_wt=True)
	assert type(m) is larch.Model
	assert 'nan_weight' in diagnosis
	assert diagnosis.nan_weight['n'].iloc[0] == 1882

	m, diagnosis = m.doctor()
	assert type(m) is larch.Model
	assert 'nan_weight' not in diagnosis
	assert len(diagnosis) == 0

def test_doctor_before_load_data():

	with raises(ValueError):
		larch.Model().doctor()




import numpy
import pandas
from .pod import Pod
from .podlist import Pods, PodsCA, EmptyPodsError
from .general import _sqz_same, _sqz, selector_len_for

import logging
from ..log import logger_name
logger = logging.getLogger(logger_name+'.data')



class DataService():
	"""A DataService is a collection of data Pod that provides data for Larch models.

	Parameters
	----------
	altids : sequence, optional
		A sequence of alternative identifiers (typically integers).
	altnames : sequence, optional
		A sequence of alternative names (typically strings). If given it must
		have the same length as the altids.
	pods : iterable of larch.Pod, optional
		Include these :class:`Pod` in the DataService. The pods can also be given as
		positional arguments to the constructor.
	"""


	def __init__(self, *args, altids=None, altnames=None, pods=None, broadcastable=True, selector=None):

		self._master_n_cases = None
		self._master_altids = [] if altids is None else altids
		self._master_altnames = [] if altnames is None else altnames

		self._pods_idco = Pods()
		self._pods_idca = PodsCA(n_alts= -1 if altids is None else len(altids) )
		self._pods_idce = Pods()

		self._pod_library = {}

		for pod in args:
			if pod is not None:
				self.add_pod(pod, broadcastable=broadcastable)

		if pods is not None:
			for pod in pods:
				self.add_pod(pod, broadcastable=broadcastable)

		self._default_selector = selector

	@property
	def selector(self):
		return self._default_selector

	@selector.setter
	def selector(self, value):
		if value is None:
			self._default_selector = None
		else:
			if not isinstance(value, (str, slice, numpy.ndarray, list, tuple)):
				raise TypeError
			if isinstance(value, str):
				self._default_selector = None
				value = self.array_idco(value, dtype=bool).squeeze()
				self._default_selector = value
			elif isinstance(value, (tuple, list)):
				self._default_selector = None
				value = self.array_idco(*value, dtype=bool).all(axis=1)
				self._default_selector = value
			else:
				if len(value)==0:
					self._default_selector = None
				else:
					self._default_selector = value

	def __getattr__(self, item):
		if item in self._pod_library:
			return self._pod_library[item]
		raise AttributeError( f"DataService object has no attribute '{item}'")

	def library_keys(self):
		return self._pod_library.keys()

	@property
	def n_cases(self):
		"""The number of cases represented by this DataService."""
		return self._master_n_cases

	@property
	def n_alts(self):
		"""The number of alternatives represented by this DataService."""
		return len(self._master_altids)

	def add_pod(self, pod:Pod, *, broadcastable=True):
		if pod.podtype in ('idco', 'idrc'):
			if self._master_n_cases is None:
				self._master_n_cases = pod.n_cases
			elif self._master_n_cases != pod.n_cases:
				raise ValueError(
					f"incompatible n_cases, have {self._master_n_cases} adding {pod.n_cases} in pod {pod!r}")
			self._pods_idco.append(pod)
			if broadcastable:
				try:
					pod_ca = pod.as_idca( )
				except AttributeError:
					pass
				else:
					self._pods_idca.append(pod_ca)

		elif pod.podtype in ('idca', 'idcs', 'idga', 'id0a'):
			if self._master_n_cases is None:
				self._master_n_cases = pod.n_cases
			elif self._master_n_cases != pod.n_cases:
				raise ValueError(
					f"incompatible n_cases, have {self._master_n_cases} adding {pod.n_cases} in pod {pod!r}")
			self._pods_idca.append(pod)
		elif pod.podtype=='idce':
			if self._master_n_cases is None:
				self._master_n_cases = pod.n_cases
			elif self._master_n_cases != pod.n_cases:
				raise ValueError(
					f"incompatible n_cases, have {self._master_n_cases} adding {pod.n_cases} in pod {pod!r}")
			self._pods_idce.append(pod)
		else:
			raise TypeError(f'podtype {pod.podtype} unknown')

		self._pod_library[pod.ident] = pod

		self._cleanup_trailing_dims()

	def _cleanup_trailing_dims(self):
		for pod in self._pods_idca:
			if pod.shape[-1] == -1:
				pod.trailing_dim = self._pods_idca.shape[-1]

	def set_alternatives(self, altids=None, altnames=None):
		if altids is not None:
			self._master_altids = altids
		if altnames is not None:
			self._master_altnames = altnames

	def alternative_codes(self):
		return self._master_altids

	def alternative_names(self):
		return self._master_altnames

	def alternative_name_dict(self):
		return {c:n for c,n in zip(self._master_altids, self._master_altnames)}

	@property
	def alternatives(self):
		"""A list of 2-tuples, each giving the id and name of an alternative"""
		return [(c,n) for c,n in zip(self._master_altids, self._master_altnames)]

	def vars_idco(self):
		return self._pods_idco.names()

	def vars_idca(self):
		return self._pods_idca.names()

	def _get_dims_of_command(self, cmd_as_utf8):
		from tokenize import tokenize, untokenize, NAME, OP, STRING
		from io import BytesIO
		g = tokenize(BytesIO(cmd_as_utf8).readline)
		# identify needed dims
		dims = 1
		for toknum, tokval, _, _, _ in g:
			if toknum == NAME and (tokval in self._pods_idca.names()):
				dims = 2
				break
		return dims


	def caseindexes(self, selector=None):
		if selector is None:
			selector = self._default_selector
		if selector is None:
			return numpy.arange(self._master_n_cases)
		else:
			return numpy.arange(self._master_n_cases)[selector]

	def array_idco(self, *vars, dtype=numpy.float64, selector=None, strip_nan=True):
		"""Extract a set of idco values into a new array.

		Parameters
		----------
		vars : tuple of str
			A tuple (or other iterable) giving the expressions (often column names, but any valid
			SQLite expression works) to extract as :ref:`idco` format variables.
		selector : None or slice
			If given, use this to slice the caseids used to build
			the array.
		strip_nan : bool
			If True (the default) then NAN values are converted to 0, and INF
			values are converted to the largest magnitude real number representable
			in the selected dtype.

		Other Parameters
		----------------
		dtype : str or dtype
			Describe the data type you would like the output array to adopt, probably
			numpy.int64, numpy.float64, or numpy.bool.

		Returns
		-------
		data : ndarray
			An array with specified dtype, of shape (n_cases,len(vars)).
		"""
		if selector is None:
			selector = self._default_selector
		result = self._pods_idco.get_data_items(vars, selector=selector, dtype=dtype)
		if strip_nan:
			result = numpy.nan_to_num(result)
		return result

	def dataframe_idco(self, *vars, selector=None, **kwargs):
		"""Extract a set of idco values into a new dataframe.

		Parameters
		----------
		vars : tuple of str
			A tuple (or other iterable) giving the expressions (often column names, but any valid
			SQLite expression works) to extract as :ref:`idco` format variables.
		selector : None or slice
			If given, use this to slice the caseids used to build
			the array.
		strip_nan : bool
			If True (the default) then NAN values are converted to 0, and INF
			values are converted to the largest magnitude real number representable
			in the selected dtype.

		Other Parameters
		----------------
		dtype : str or dtype
			Describe the data type you would like the output array to adopt, probably
			numpy.int64, numpy.float64, or numpy.bool.

		Returns
		-------
		data : pandas.DataFrame
			A DataFrame with data of specified dtype
		"""
		if selector is None:
			selector = self._default_selector
		d = self.array_idco(*vars, selector=selector, **kwargs)
		return pandas.DataFrame(
			data=d,
			columns=vars,
			index=self.caseindexes(selector=selector),
		)


	def load_idco(self, *names, arr=None, dtype=numpy.float64, selector=None, strip_nan=True, mask_pattern=0, mask_names=None, log=None, fallback_to_idce=False):
		"""Extract a set of idco values.

		Parameters
		----------
		names : tuple of str
			A tuple (or other iterable) giving the expressions (often column names, but any valid
			SQLite expression works) to extract as :ref:`idco` format variables.
		arr : array-like or None
			The array into which the data shall be loaded.
		dtype : str or dtype
			Describe the data type you would like the output array to adopt, probably
			numpy.int64, numpy.float64, or numpy.bool.
		selector : None or slice
			If given, use this to slice the caseids used to build
			the array.
		strip_nan : bool
			If True (the default) then NAN values are converted to 0, and INF
			values are converted to the largest magnitude real number representable
			in the selected dtype.
		mask_pattern : int, optional
		mask_names : array-like, optional
			An array of int32 of the same shape as `names`.  For each item, if the
			mask_names value bitwise-and the mask_pattern evaluates
			as True, then the data will be skipped. Note the default mask_pattern is 0 so
			even if mask_names is given, everything will be loaded unless the the mask_pattern
			is set to some other value. If mask_pattern is not zero but mask_names is not given,
			the mask_names will be inferred from get_durable_mask values of the component pods.
			This is an optimization tool
			for reloading data that may not have changed (e.g. for logsum generation).

		Other Parameters
		----------------

		Returns
		-------
		data : ndarray
			An array with specified dtype, of shape (n_cases,len(vars)).
		"""
		if selector is None:
			selector = self._default_selector
		try:
			if arr is None:
				arr = numpy.zeros(self.idco.shape_result(selector, names), dtype=dtype)
			else:
				if _sqz(arr.shape) != _sqz(self.idco.shape_result(selector, names)):
					import warnings
					newshape = self.idco.shape_result(selector, names)
					warnings.warn(f"load_idco: injection array is not correctly sized, re-initializing from {arr.shape} to {newshape}")
					raise ValueError
					arr = numpy.zeros(newshape, dtype=dtype)
		except EmptyPodsError:
			if fallback_to_idce:
				raise NotImplementedError
			else:
				return None

		if len(names)==0:
			return arr

		if mask_pattern != 0 and mask_names is None:
			mask_names = self.idco.get_data_masks(names)


		result = self.idco.load_data_items(names, result=arr, selector=selector, dtype=dtype,
												 mask_pattern=mask_pattern, mask_names=mask_names, log=log)
		if strip_nan:
			result = numpy.nan_to_num(result)
		return result

	def load_idce(self, *names, arr=None, dtype=numpy.float64, selector=None, strip_nan=True, mask_pattern=0, mask_names=None, log=None):
		"""Extract a set of idce values, with indexes.

		Parameters
		----------
		names : tuple of str
			A tuple (or other iterable) giving the expressions (often column names, but any valid
			SQLite expression works) to extract as :ref:`idco` format variables.
		arr : array-like or None
			The array into which the data shall be loaded.
		dtype : str or dtype
			Describe the data type you would like the output array to adopt, probably
			numpy.int64, numpy.float64, or numpy.bool.
		selector : None or slice
			If given, use this to slice the caseids used to build
			the array.
		strip_nan : bool
			If True (the default) then NAN values are converted to 0, and INF
			values are converted to the largest magnitude real number representable
			in the selected dtype.
		mask_pattern : int, optional
		mask_names : array-like, optional
			An array of int32 of the same shape as `names`.  For each item, if the
			mask_names value bitwise-and the mask_pattern evaluates
			as True, then the data will be skipped. Note the default mask_pattern is 0 so
			even if mask_names is given, everything will be loaded unless the the mask_pattern
			is set to some other value. If mask_pattern is not zero but mask_names is not given,
			the mask_names will be inferred from get_durable_mask values of the component pods.
			This is an optimization tool
			for reloading data that may not have changed (e.g. for logsum generation).

		Other Parameters
		----------------

		Returns
		-------
		data : idce_arrays
		"""
		if selector is None:
			selector = self._default_selector
		try:
			if arr is None:
				arr = numpy.zeros(self.idce.shape_result(selector, names, use_metashape=True), dtype=dtype)
			else:
				if _sqz(arr.shape) != _sqz(self.idce.shape_result(selector, names, use_metashape=True)):
					import warnings
					newshape = self.idce.shape_result(selector, names)
					warnings.warn(f"load_idce: injection array is not correctly sized, re-initializing from {arr.shape} to {newshape}")
					raise ValueError
					# arr = numpy.zeros(newshape, dtype=dtype)
		except EmptyPodsError:
			return None

		if len(names)==0:
			return None

		if mask_pattern != 0 and mask_names is None:
			mask_names = self.idce.get_data_masks(names)

		caseindexes, altindexes = self.idce.load_casealt_indexes()
		result = self.idce.load_data_items(
			names,
			result=arr,
			selector=selector,
			dtype=dtype,
			mask_pattern=mask_pattern,
			mask_names=mask_names,
			log=log,
			use_metashape=True,
		)
		if strip_nan:
			result = numpy.nan_to_num(result)

		from ..model.idce_tools import idce_arrays
		return idce_arrays(caseindexes, altindexes, result)


	def array_idca(self, *vars, dtype=numpy.float64, selector=None, strip_nan=True):
		"""Extract a set of idca values.

		Parameters
		----------
		vars : tuple of str
			A tuple (or other iterable) giving the expressions (often column names, but any valid
			SQLite expression works) to extract as :ref:`idco` format variables.
		selector : None or slice
			If given, use this to slice the caseids used to build
			the array.
		strip_nan : bool
			If True (the default) then NAN values are converted to 0, and INF
			values are converted to the largest magnitude real number representable
			in the selected dtype.

		Other Parameters
		----------------
		dtype : str or dtype
			Describe the data type you would like the output array to adopt, probably
			numpy.int64, numpy.float64, or numpy.bool.

		Returns
		-------
		data : ndarray
			An array with specified dtype, of shape (n_cases,len(vars)).
		"""
		if selector is None:
			selector = self._default_selector
		result = self._pods_idca.get_data_items(vars, selector=selector, dtype=dtype)
		if strip_nan:
			result = numpy.nan_to_num(result)
		return result

	def dataframe_idca(self, *vars, selector=None, **kwargs):
		"""Extract a set of idca values into a new dataframe.

		Parameters
		----------
		vars : tuple of str
			A tuple (or other iterable) giving the expressions (often column names, but any valid
			SQLite expression works) to extract as :ref:`idca` format variables.
		selector : None or slice
			If given, use this to slice the caseids used to build
			the array.
		strip_nan : bool
			If True (the default) then NAN values are converted to 0, and INF
			values are converted to the largest magnitude real number representable
			in the selected dtype.

		Other Parameters
		----------------
		dtype : str or dtype
			Describe the data type you would like the output array to adopt, probably
			numpy.int64, numpy.float64, or numpy.bool.

		Returns
		-------
		data : pandas.DataFrame
			A DataFrame with data of specified dtype
		"""
		if selector is None:
			selector = self._default_selector
		d = self.array_idca(*vars, selector=selector, **kwargs)
		d = d.reshape(d.shape[0]*d.shape[1], -1)

		rows = pandas.MultiIndex.from_product(
			[self.caseindexes(selector=selector), self.alternative_codes(), ],
			names=['case', 'alt', ]
		)

		return pandas.DataFrame(
			data=d,
			columns=vars,
			index=rows,
		)

	def load_idca(self, *names, arr=None, dtype=numpy.float64, selector=None, strip_nan=True, mask_pattern=0, mask_names=None, log=None, fallback_to_idce=False):
		"""Extract a set of idca values.

		Parameters
		----------
		names : tuple of str
			A tuple (or other iterable) giving the expressions (often column names, but any valid
			SQLite expression works) to extract as :ref:`idco` format variables.
		arr : array-like or None
			The array into which the data shall be loaded.
		dtype : str or dtype
			Describe the data type you would like the output array to adopt if not given, probably
			numpy.int64, numpy.float64, or numpy.bool.
		selector : None or slice
			If given, use this to slice the caseids used to build
			the array.
		strip_nan : bool
			If True (the default) then NAN values are converted to 0, and INF
			values are converted to the largest magnitude real number representable
			in the selected dtype.
		mask_pattern : int, optional
		mask_names : array-like, optional
			An array of int32 of the same shape as `names`.  For each item, if the
			mask_names value bitwise-and the mask_pattern evaluates
			as True, then the data will be skipped. Note the default mask_pattern is 0 so
			even if mask_names is given, everything will be loaded unless the the mask_pattern
			is set to some other value. If mask_pattern is not zero but mask_names is not given,
			the mask_names will be inferred from get_durable_mask values of the component pods.
			This is an optimization tool
			for reloading data that may not have changed (e.g. for logsum generation).

		Returns
		-------
		data : ndarray
			An array with specified dtype, of shape (n_cases,len(vars)).
		"""
		if selector is None:
			selector = self._default_selector
		try:
			if arr is None:
				arr = numpy.zeros(self.idca.shape_result(selector, names), dtype=dtype)
			else:
				out_shape = self.idca.shape_result(selector, names)
				if _sqz(arr.shape) != _sqz(out_shape):
					if (arr.shape[0]!=out_shape[0]) or (arr.shape[1] < out_shape[1]):
						import warnings
						newshape = self.idca.shape_result(selector, names)
						warnings.warn(f"load_idca: injection array is not correctly sized, re-initializing from {arr.shape} to {newshape}")
						raise ValueError
						arr = numpy.zeros(newshape, dtype=dtype)
		except EmptyPodsError:
			if fallback_to_idce:
				temp = self.load_idce(*names, arr=None, dtype=dtype, selector=selector, strip_nan=strip_nan,
									  mask_pattern=mask_pattern, mask_names=mask_names, log=log)
				return temp.as_idca()
			else:
				return None

		if len(names)==0:
			return arr

		if mask_pattern != 0 and mask_names is None:
			mask_names = self.idca.get_data_masks(names)

		try:
			result = self.idca.load_data_items(names, result=arr, selector=selector, dtype=dtype,
												 mask_pattern=mask_pattern, mask_names=mask_names, log=log)
		except NameError:
			if fallback_to_idce:
				temp = self.load_idce(*names, arr=None, dtype=dtype, selector=selector, strip_nan=strip_nan,
									  mask_pattern=mask_pattern, mask_names=mask_names, log=log)
				result[:] = temp.as_idca()
			else:
				raise

		if strip_nan:
			result = numpy.nan_to_num(result)
		return result

	make_idco = dataframe_idco
	make_idca = dataframe_idca

	def __repr__(self):
		s = "<larch.DataService>"
		if len(self._pods_idco):
			s += "\n | idco:"
			for dat in self._pods_idco:
				s += "\n |   "
				s += repr(dat).replace("\n","\n |   ")
		if len(self._pods_idca):
			s += "\n | idca:"
			for dat in self._pods_idca:
				s += "\n |   "
				s += repr(dat).replace("\n","\n |   ")
		if len(self._pods_idce):
			s += "\n | idce:"
			for dat in self._pods_idce:
				s += "\n |   "
				s += repr(dat).replace("\n","\n |   ")
		return s

	@property
	def idco(self):
		return self._pods_idco

	@property
	def idce(self):
		return self._pods_idce

	@property
	def idca(self):
		return self._pods_idca

	def idca_reset(self, pull_idco=False):
		self.idca.clear()
		self.idca.n_alts = self.n_alts
		if pull_idco:
			for ident, pod in self._pod_library.items():
				try:
					pod_ca = pod.as_idca( )
				except AttributeError:
					pass
				else:
					self._pods_idca.append(pod_ca)


	@classmethod
	def from_CE_and_SA(cls, ce, sa, *, co=None):
		from .h5 import H5PodCO
		return cls(
			ce,
			co,
			H5PodCO(shape=(ce.shape[0],)),
			altids=sa.altcodes,
			altnames=sa.altnames,
			broadcastable=False,
		)


	def make_dataframes(self, req_data, *, selector=None, float_dtype=numpy.float64, log_warnings=True):
		"""Create a DataFrames object that will satisfy a data request.

		Parameters
		----------
		req_data : Dict or str
			The requested data. The keys for this dictionary may include {'ca', 'co',
			'choice_ca', 'choice_co', 'weight_co', 'avail_ca', 'standardize'}.
			Currently, the keys {'choice_co_code', 'avail_co'} are not implemented and
			will raise an error.
			Other keys are silently ignored.
		selector : array-like[bool] or slice, optional
			If given, the selector filters the cases. This argument can only be given
			as a keyword argument.
		float_dtype : dtype, default float64
			The dtype to use for all float-type arrays.  Note that the availability
			arrays are always returned as int8 regardless of the float type.
			This argument can only be given
			as a keyword argument.
		log_warnings : bool, default True
			No effect.

		Returns
		-------
		DataFrames
			This object should satisfy the request.
		"""

		if isinstance(req_data, str):
			from ..util import Dict
			import textwrap
			req_data = Dict.load(textwrap.dedent(req_data))

		if 'ca' in req_data:
			logger.info("Loading `ca` data...")
			df_ca = self.dataframe_idca(*req_data['ca'], dtype=float_dtype, selector=selector)
		else:
			df_ca = None

		if 'co' in req_data:
			logger.info("Loading `co` data...")
			df_co = self.dataframe_idco(*req_data['co'], dtype=float_dtype, selector=selector)
		else:
			df_co = None

		if 'choice_ca' in req_data:
			logger.info("Loading `choice_ca` data...")
			df_ch = self.dataframe_idca(req_data['choice_ca'], dtype=float_dtype, selector=selector)
		elif 'choice_co' in req_data:
			logger.info("Loading `choice_co` data...")
			alts = self.alternative_codes()
			cols = [req_data['choice_co'].get(a, '0') for a in alts]
			df_ch = self.dataframe_idco(*cols, dtype=float_dtype, selector=selector)
			df_ch.columns = alts
		elif 'choice_co_code' in req_data:
			logger.info("Loading `choice_co_code` data...")
			alts = self.alternative_codes()
			df_ch_code = self.dataframe_idco(req_data['choice_co_code'], dtype=int, selector=selector)
			df_ch = pandas.DataFrame(0, columns=alts, index=df_ch_code.index, dtype=float_dtype)
			for c in df_ch.columns:
				df_ch.loc[:,c] = (df_ch_code==c).astype(float_dtype)
		else:
			df_ch = None

		if 'weight_co' in req_data:
			logger.info("Loading `weight_co` data...")
			df_wt = self.dataframe_idco(req_data['weight_co'], dtype=float_dtype, selector=selector)
		else:
			df_wt = None

		if 'avail_ca' in req_data:
			logger.info("Loading `avail_ca` data...")
			df_av = self.dataframe_idca(req_data['avail_ca'], dtype=numpy.int8, selector=selector)
		elif 'avail_co' in req_data:
			raise NotImplementedError('avail_co')
		else:
			df_av = None

		from ..dataframes import DataFrames

		logger.info("Combining into DataFrames structure...")
		result = DataFrames(
			co=df_co,
			ca=df_ca,
			av=df_av,
			ch=df_ch,
			wt=df_wt,
			alt_codes=self.alternative_codes(),
			alt_names=self.alternative_names(),
		)

		if 'standardize' in req_data and req_data.standardize:
			logger.info("Standardizing data...")
			result.standardize()

		logger.info("Completed make_dataframes.")
		return result


	def validate_dataservice(self, req_data):
		"""
		Check if an object is a sufficient dataservice.
		"""

		if isinstance(req_data, str):
			from ..util import Dict
			import textwrap
			req_data = Dict.load(textwrap.dedent(req_data))

		missing_methods = set()

		if 'ca' in req_data:
			if not hasattr(self, 'dataframe_idca'):
				missing_methods.add('dataframe_idca')

		if 'co' in req_data:
			if not hasattr(self, 'dataframe_idco'):
				missing_methods.add('dataframe_idco')

		if 'choice_ca' in req_data:
			if not hasattr(self, 'dataframe_idca'):
				missing_methods.add('dataframe_idca')
		elif 'choice_co' in req_data:
			if not hasattr(self, 'dataframe_idco'):
				missing_methods.add('dataframe_idco')
		elif 'choice_co_code' in req_data:
			if not hasattr(self, 'dataframe_idco'):
				missing_methods.add('dataframe_idco')

		if 'weight_co' in req_data:
			if not hasattr(self, 'dataframe_idco'):
				missing_methods.add('dataframe_idco')

		if 'avail_ca' in req_data:
			if not hasattr(self, 'dataframe_idca'):
				missing_methods.add('dataframe_idca')
		elif 'avail_co' in req_data:
			raise NotImplementedError('avail_co')

		if len(missing_methods)>0:
			raise ValueError('dataservice is missing '+", ".join(missing_methods))




import pandas
import os
import gzip
import pickle
import base64
import io
import numpy
from pytest import approx, raises

from larch.data_warehouse import example_file

from larch import DataFrames, DBF, data_warehouse


def test_dfs_info():

	from larch.data_warehouse import example_file
	df = pandas.read_csv(example_file("MTCwork.csv.gz"))
	df.set_index(['casenum', 'altnum'], inplace=True)

	ds = DataFrames(df)

	s = io.StringIO()

	ds.info(out=s)

	assert s.getvalue() == (
		'larch.DataFrames:  (not computation-ready)\n'
		'  n_cases: 5029\n'
		'  n_alts: 6\n'
		'  data_ce: 36 variables, 22033 rows\n'
		'  data_co: <not populated>\n'
		'  data_av: <populated>\n')

	s = io.StringIO()
	ds.info(out=s, verbose=True)

	assert s.getvalue() == (
		'larch.DataFrames:  (not computation-ready)\n  n_cases: 5029\n  n_alts: 6\n  data_ce: 22033 rows\n'
		'    - chose    (22033 non-null int64)\n    - ivtt     (22033 non-null float64)\n'
		'    - ovtt     (22033 non-null float64)\n    - tottime  (22033 non-null float64)\n'
		'    - totcost  (22033 non-null float64)\n    - hhid     (22033 non-null int64)\n'
		'    - perid    (22033 non-null int64)\n    - numalts  (22033 non-null int64)\n'
		'    - dist     (22033 non-null float64)\n    - wkzone   (22033 non-null int64)\n'
		'    - hmzone   (22033 non-null int64)\n    - rspopden (22033 non-null float64)\n'
		'    - rsempden (22033 non-null float64)\n    - wkpopden (22033 non-null float64)\n'
		'    - wkempden (22033 non-null float64)\n    - vehavdum (22033 non-null int64)\n'
		'    - femdum   (22033 non-null int64)\n    - age      (22033 non-null int64)\n'
		'    - drlicdum (22033 non-null int64)\n    - noncadum (22033 non-null int64)\n'
		'    - numveh   (22033 non-null int64)\n    - hhsize   (22033 non-null int64)\n'
		'    - hhinc    (22033 non-null float64)\n    - famtype  (22033 non-null int64)\n'
		'    - hhowndum (22033 non-null int64)\n    - numemphh (22033 non-null int64)\n'
		'    - numadlt  (22033 non-null int64)\n    - nmlt5    (22033 non-null int64)\n'
		'    - nm5to11  (22033 non-null int64)\n    - nm12to16 (22033 non-null int64)\n'
		'    - wkccbd   (22033 non-null int64)\n    - wknccbd  (22033 non-null int64)\n'
		'    - corredis (22033 non-null int64)\n    - vehbywrk (22033 non-null float64)\n'
		'    - vocc     (22033 non-null int64)\n    - wgt      (22033 non-null int64)\n'
		'  data_co: <not populated>\n  data_av: <populated>\n')

	assert not ds.computational
	assert not ds.is_computational_ready()
	ds.computational = True
	assert ds.is_computational_ready()
	assert ds.computational
	s = io.StringIO()

	ds.info(out=s)

	assert s.getvalue() == (
		'larch.DataFrames:\n'
		'  n_cases: 5029\n'
		'  n_alts: 6\n'
		'  data_ce: 36 variables, 22033 rows\n'
		'  data_co: <not populated>\n'
		'  data_av: <populated>\n')

def test_service_idco():

	df = pandas.read_csv(example_file("MTCwork.csv.gz"))
	df.set_index(['casenum', 'altnum'], inplace=True)

	dfs = DataFrames(df, crack=True)
	check1 = dfs.make_idco('1')
	assert (check1 == 1).shape == (5029, 1)
	assert numpy.all(check1 == 1)

	check2 = dfs.make_idco('age')
	assert check2.shape == (5029, 1)
	assert numpy.all(check2.iloc[:5,0] == [35, 40, 28, 34, 43])
	assert numpy.all(check2.iloc[-5:,0] == [58, 33, 34, 35, 37])

	check3 = dfs.make_idco('age', '1')
	assert check3.shape == (5029, 2)



def test_dbf_reader():


	q = DBF(data_warehouse.example_file('US-STATES.dbf'))

	assert q.fieldnames() == [
		'STATEFP',
		'STATENS',
		'AFFGEOID',
		'GEOID',
		'STUSPS',
		'NAME',
		'LSAD',
		'ALAND',
		'AWATER',
	]

	df = q.load_dataframe(preserve_order=False, strip_whitespace=False)

	correct = b'ABzY8<${%50{^v}dwdkt^}s_!9zhTV6cBJlk*FB=HM6sbNXP@i=EWu?0gUNnGs!O5?B<-EC4?Fg6%a&4#cB}W7GJGeA4P=<ilV4h#' \
			  b'EMT8pA{dKpS8czTHD?`bMMW~w7-9T?d0Q!^F8O@x#!+9=iWIJCzl)*>(o;%dZr+jMx&q=^$sJiSU2@jiy-HWl`x<Oa`PLG&dJGXY9' \
			  b'TGq|B^(?G%%$nGX))bN^dhTlQ%n=j<r|VS{@7}LmZvo2!rw+*R>eSj7&kZo-y*8Qbt!(8MWlW;QzZD>oKGUhUPa6BongclNU>6J37' \
			  b'1V{vFw*na~mS?z|>6V(C=I#B9vWgJE{BJTkP^3dclpR>AP10nIRC4ICSVk@=AUk)f;LIM%Vn?CvBHIKCm5II7ZMR3s-dU=56p!kGN' \
			  b'XQ4Ir#^&&W-va)jCKgf^uPvm?PPHbM=2xIMd6_yuP7Q?uPL6qwYYfGXqzF{y?YHEp6S`8=lX_YlFp<xK^SXw%#q^hhKCR$_IJjoiP' \
			  b'rglM1bq$<st-Z3aq69n^me&*(!zqRdlbbC)MeZG(TA5u9r$u2({NG2EtkS?lj0DwGgDkSA_rTQrcwT&1+=@<qoP_3#D4a<{qpgCo;' \
			  b'&wC=8cegJads36cr<3fIsK!NV}iGN?d+Yjk^OMYW%4K{2QqmylY^KX%;XRzhccPR<S-_WVRAT=BbYpv$&pMR$K>%$j$(2&lVg}Xfy' \
			  b'on@9LwZ5CdV^*5|a~{oXF%PCQoM4!{jMUPG<5{CQoB>3X`WZIhDzLCeL8<OeW7_avGCoGg-joIZS$)^fBpYGQebzNrg$3$q<v^Gr`' \
			  b'{+w{j{Fg`m+0N<L2$m8MB33K~t5+=#Xs!tsK&vo9exw`<Ea>u4<Qs9idNCP~lpy(@}|zBxZvr?HvtTlZZfo$vNLSIngMv%B{znj1?' \
			  b';@1I6<WmdM+y5(2Q3T>k?>zc9bLz*+m8~*qvT}xHGdpV71)8ZYQ=)Jr4*aHtz{jj#Yg0}aTef=)Ysjkag2htcExOGQ_-mCg&-hZy1' \
			  b'$RD+Zf!RcUk-02QbI=&_{gg5y?|g6gW*Y0Qvu|jmvK}Y>K=1j~>%a2Q+*$N(=~f!+a|3_gK>b`3JaRSl^X}&PD`@WRZC$>8I+2YV_' \
			  b'E;&J?A|z%u4~%xH4QWus&2gUHJV$+=I{}8+#h$W>80a-T07@4I*+sWB*xM8p84@Ells5^sz++*J-xJdR4et<`_wl}=sI>^ckz6hbB' \
			  b'~r!xtr$bz7vjo6ejYso?C9Ce$UOh>j1UyoxWn-EktGy=30Im=RbQq&8a8r|8y^%&z8JaD@C8U@N_Gc!?T~Zw!}LpzIq(J58L)8Cev' \
			  b'}%ku%Pt_I-ytAEt6w-JxA{zFS_&E2VnT5hY3Gj6E0DP`NES!P@fG*&D}GShQro(=-m_UOsX)wSRE?cR=k6gT0HXe(kWCv#kB|&3V;' \
			  b'SZu(@#?X-PyRdL+9KhKtaN$oFQ(epW#ZDWTlr*ix1iEmi*S~ASq0zd8#4x-mhzGv(ns=sh**GtxUv2<1ewO_U6Qp>)fE_eskS8mN(' \
			  b'>pB<>SX+4C?!zf64}Je-1(gqMhbgq*mEZ2Hq-WAYJ9TUR@wZ}Y=>7O%_nhr?UFCNTypQ&q_VzA~#{KbF$wu1mqQ!4n>v^iB+VV5C?' \
			  b'yt+~dfMMPIGN^i=e*Zr^!iP&6<YDw`0^)}wBPz#eZ6%)-ZV5t`>*sj{EfEX{LH$qY5&84QTypxIe6d0)>fTz==IT7yxwoUz}mOdO*' \
			  b'priZ`2udqn;OqnRKHLrW^JA_!#oInw4)I>s}ue+Te%THUy!_1_g?3P@%*I4N7eY!yF5IUYKix56Wx^z&slih}fV)xeXzxut9@L8^T' \
			  b'a!f!_<&HUwe54Izlyph1lV0WZ|r;D-e^1fk9b1?p|kpuvKm7Z%#!heb9hu-FC_F0dgCjTR_gxX^|mG})lSPb^Tqpj!~~L$f_+Kg8^' \
			  b'b`JlxHKNvOypw$Ki+H6oEZbKLn_6&Wn#D)O0+n_?yh7fev5QdaJLm#AV2tua~ApjdRKzph_$k-47(*^~y_T*JqYC{;h>{V#6%&IuN' \
			  b'zECI}3MtUdRZUUIl$Oh_A^54SdcCUV)fBi$Zq=YiR>N?yq-tJRA*(*PgsWbEK=Z>&p(+8A7?;YO{cxGA2H<j84Z<omt6)%5NKUSnT' \
			  b'NPNtTQ#3XvhfPu>I;TR!mMSjs_NHB0<4p+4SV5A)*4bnVKR;Na%%vtl2rw+W}UTgNFf<`jocc74NUcU{hAVjYvooAu9MX;Trb~+AZ' \
			  b'%o6P*K8K7;fn6tdV{CGg(#OMp+HPO-v;{Lu7aN%B>n~VsrL~gF$i!Z<a%+!e*wb{;(1Zz%6oX2yW$}^QvKz*tf~8VYpphkq_>WRX^' \
			  b'M*t3kL+R#mv0sbtD3IWB(ARZR(!NN(YgR1_as;5|(B`@$Lt`@KvJ_!TV_fcxb2`rv+94Zs7k8iZfSstUi9L#M!la%&hKl3K~R_i)2' \
			  b'7>$LRJp9M9<*RLEFKRn{F1mRJKMS;g078M?MSTy*x!xDzA{VilZPdF?-c+z1Bz*7#30^1xG6}CGpA^45MqQTP+OBi<aw~+IEr^6D2' \
			  b'XB?IgJnOJ%u&cj?ocYf=EPi<2VF|)+hed%G92N~;>~A3th&>LAA6{};6nNQTQQ;MbB@DmqZy`^UR~?ohyymc|u(!X3JZN6;Zy`^de' \
			  b'U3Q$;SEQ`eDFJm#Sd>fECJZ>uqg1B!=l354oeu`am3IE?>a02c+X)`;eCfC1P2_JFnr*Mp$|TESb}iSVF|%U4vPjKJEH1?PaKv2eC' \
			  b'n_$@R=j>Dtzv+gyHv&d!fM>yk0uQ>!mMwz4R5Ym%irp(l@+b`U9_*zUB4OVO}r&53iTL<Mq-XdA;;Kub2MB>!ttY_0pety>x`vOMl' \
			  b'_@(qDPK^aHP#{>JO2A9=m>cU~|3gVjq&-_KdegaZWoF;}pHM+sJOpx`hb%~|o{Ai+T#ELg=MoK-Il<s9;39%r8)hjI4%@EE~<94<J' \
			  b'3BLpjWtY8&K3J&9OoC7{QUT^?M3085m;1G@x9L5vGx_o$|;2@3_9KvyeH5@P2=);o)2XKO51t*HNt2jw;7*7^)(9pxVupCbj9Kp$g' \
			  b'7vZUbqj(zUqC%V^xC&1fT#HjV7nfkZ;37PO^PECFlk?m%Jd1N#F-{X)j%N$5!~(&Ecn;@C4SG3O6r)dYCHe)|U_fv!202&Op(1!5s' \
			  b')8#pB)Awg&Q%c%bFMDL>6~kd@La*gID_+oT0D<)T`kV!TwjOh3tot`L~LuYP>dhN*_=zuu}G|UAr^BkFT@hDt`aO2T!?c7m*HH&Ra' \
			  b'hq0Rf6*b&&7z~g;*}Q2rI<e=VGPcC{}TfmSHvL+De=+xB;VL?S)t))?0|Rg3EA$$cF`3C+1m=_2Swxq`$&xXrb{*>dktG?s4-67fR' \
			  b'kT?e04B;3DA-6U);nm)Ewq&s)2m%IFzay<Nb)73s-DYZ_At7ng^P!dp6-M!hBN>e?>s^Oi~(F*6a%n(ldP;@*mlq~4{Y;p*CcBD^I' \
			  b'sCMII0C*A5HQO$NV|M%}8>fGBLBaLkdmv=xj_coWrF_Cb2cgMK5vKBp__PFe=MR-dl6RC8<-Q8^nZ>g4q?)J&mD!iq;Tt1VDw>IuA' \
			  b';`%8`yv4b<3LU$XdaA`$ZwcWoK~Anr+H^CVxTMcpCPV-0>~wF_VY~2_PMNOfWRiQUG?wZuZdVE&ecrk}^9;-w-7b1d$=)iAWr>)J#' \
			  b'}}rBw=|k@PqE(aVp?J+_g2-OB*(J51tkLaR&AtGneOCL-NiZxk$bC2XGtWBbWD?HflK}{)8CuR_s!(q7MOb6^=Ckqd#f|BgB&OBdR' \
			  b'xl9)f*YpQ<or>kb7x@UBX*;x+9Tlb9Zga#9N2;?T4Q7)aV^)w{o&u?6<6eX4(^V^XS6m+*>Jm=}5Tz5Wt^`y2MDC*;sqGtKKf+-pb' \
			  b'S2M8@SB4}1EmphU`Wdv{;l=Pi@bWASXpFip1#JzUXO1;yftHnIcUT-zmb&6Uogt8Y2DvaeETMz?=sf|ts5$y~joGZRm^_BxE0$=(*' \
			  b'GvE9|RUEaS6a{1Z4N>oAdgxjyC9$YQFq0yF3x$JEX_ZE$(TMSQG#?7}Xyn=hHHRu~v#&GlW!nJ+gmbv|E>cMrQ3Uaq3vAta>>Jpvo' \
			  b'>r|WRW~ShJ;VqG&-|QybOiR2<{GKbTceuXauI~H&WCy8Uy9_-^j+At($BY{u>wGaV<*CrodaT<U<o}&ac#WvcQwh^(@zj{QX=GR#`' \
			  b'=5^fFWxrr|2N?E;<XZAC-L<XZ<P23iGL>XjS}A^aj(RiB)(bV%@W@t@vRcyCh_eO-y!jx65l29-4g#?;w=*2Bk{cw-zV|?5<ejEFC' \
			  b'_k@#1BgRki-v5{40qck@!)GACvfTiGMBeR*9dG_(_SMl6afM+a>;u#7|4SL*kthKO^z867Q1uIf<W_c(=qaNc<vU!^AyAG#ich5|R' \
			  b'1w!SNwI__BrkOLu&QsPPeN@V7)p@m1PLzhz#9uhE7vYa4N|b-}p!crxDWbUf>0XxwLA@CNN=jr+U)<G#tp-EVb!i;X*Y6~4{KeaE`' \
			  b'sr1*sR#Q3E6$#GBo6yo9CJ`e9%m%dL|(t`)+;^+sT-U45+z=t2uZ>sS_I)?S-zrM`K;6Xa}Agfz5{Zot#enc#d_%ZF)gP&NvY3ouw' \
			  b'nI-){rTsPQi>8h?Y$BIr@H1kIHD?n^Gm*;VWs-@Qk%ylXV<ab%8yT<$e{W6e3-WKa2M^I9mssP>TtY8hVs*`p45eMaw7PypyBgVyi' \
			  b'C@#jS<`JYyF2k4x~8G*uQ{y0K*2w7Q&Wdw>eP<kT1z-g2Wc+&Cvbvda<?)700'

	correct_df = pickle.loads(gzip.decompress(base64.b85decode(correct)))
	pandas.testing.assert_frame_equal(correct_df, df)

	df_o = q.load_dataframe(preserve_order=True, strip_whitespace=False)
	pandas.testing.assert_frame_equal(correct_df[[
		'STATEFP',
		'STATENS',
		'AFFGEOID',
		'GEOID',
		'STUSPS',
		'NAME',
		'LSAD',
		'ALAND',
		'AWATER',
	]], df_o)

	df_s = q.load_dataframe(preserve_order=False, strip_whitespace=True)
	correct_df.NAME = correct_df.NAME.str.strip()
	pandas.testing.assert_frame_equal(correct_df, df_s)

def test_repeated_splitting():
	df = pandas.read_csv(example_file("MTCwork.csv.gz"))
	df.set_index(['casenum', 'altnum'], inplace=True)

	dfs = DataFrames(df, crack=False)
	d1, d2 = dfs.split([80, 20])
	assert d1.n_cases == 4024
	assert d2.n_cases == 1005
	d11, d12 = d1.split([50, 50])
	assert d11.n_cases == 2012
	assert d12.n_cases == 2012

	dfs = DataFrames(df, crack=False)
	d1, d2 = dfs.split([80, 20], method='shuffle')
	assert d1.n_cases == 4024
	assert d2.n_cases == 1005
	d11, d12 = d1.split([50, 50])
	assert d11.n_cases == 2012
	assert d12.n_cases == 2012

	dfs = DataFrames(df, crack=True)
	d1, d2 = dfs.split([80, 20])
	assert d1.n_cases == 4024
	assert d2.n_cases == 1005
	d11, d12 = d1.split([50, 50])
	assert d11.n_cases == 2012
	assert d12.n_cases == 2012

	dfs = DataFrames(df, crack=True)
	d1, d2 = dfs.split([80, 20], method='shuffle')
	assert d1.n_cases == 4024
	assert d2.n_cases == 1005
	d11, d12 = d1.split([50, 50])
	assert d11.n_cases == 2012
	assert d12.n_cases == 2012


def test_co_only():

	x_co = pandas.DataFrame(
		numpy.random.random([20, 3]),
		columns=['Aa', 'Bb', 'Cc'],
	)
	x_co.index.name = 'caseid'

	with raises(ValueError):
		# missing altcodes
		DataFrames(
			co=x_co,
			av=True,
		)

	with raises(ValueError):
		# trying to use ca
		DataFrames(
			ca=x_co,
			av=True,
		)

	d = DataFrames(
		co=x_co,
		av=True,
		alt_codes=[1, 2, 3, 4, 5, 6],
	)
	assert d.n_alts == 6
	assert d.n_cases == 20
	assert d.data_av.shape == (20, 6)
	all(d.data_av.dtypes == numpy.int8)
	all(d.data_av == 1)


def test_ca_initialization():
	cax = pandas.DataFrame({
		'caseid': [1, 1, 1, 2, 2, 2],
		'altid_bad': ['aa', 'bb', 'cc', 'aa', 'bb', 'cc'],
		'altid_good': [1, 2, 3, 1, 2, 3],
		'altid_str': ['1', '2', '3', '1', '2', '3'],
		'buggers': [1.2, 3.4, 5.6, 7.8, 9.0, 5.5],
		'baggers': [22, 33, 44, 55, 66, 77],
	})

	import pytest

	with pytest.raises(ValueError):
		d = DataFrames(
			ca=cax.set_index(['caseid', 'altid_bad']),
		)

	d = DataFrames(
		ca=cax.set_index(['caseid', 'altid_good']),
	)
	assert all(d.alternative_codes() == [1, 2, 3])

	d = DataFrames(
		ca=cax.set_index(['caseid', 'altid_str']),
	)
	assert all(d.alternative_codes() == [1, 2, 3])


def test_dfs_init_ca():
	from larch.data_warehouse import example_file

	df = pandas.read_csv(
		example_file("MTCwork.csv.gz"),
		index_col=['casenum', 'altnum']
	)

	d0 = DataFrames(ca=df, crack=True, ch='chose', wt_name='wgt')
	assert d0.data_wt is not None
	assert d0.data_wt.columns == 'wgt'
	assert d0.data_ch is not None
	assert d0.data_ch.shape == (5029, 6)
	assert d0.data_av is not None
	assert d0.data_av.shape == (5029, 6)

	d1 = DataFrames(ca=df, crack=True)
	assert d1.data_wt is None
	assert d1.data_ch is None
	assert d1.data_av is not None
	assert d1.data_av.shape == (5029, 6)

	d2 = DataFrames(df)
	assert d2.data_wt is None
	assert d2.data_ch is None
	assert d2.data_av is not None
	assert d2.data_av.shape == (5029, 6)
	assert d2.data_co is None
	assert d2.data_ca is None
	assert d2.data_ce is not None
	assert d2.data_ce.shape == (22033, 36)

	d3 = DataFrames(ca=df, crack=True, ch=df.chose, wt_name='wgt')
	assert d3.data_wt is not None
	assert d3.data_wt.columns == 'wgt'
	assert d3.data_ch is not None
	assert d3.data_ch.shape == (5029, 6)
	assert d3.data_av is not None
	assert d3.data_av.shape == (5029, 6)
	assert pandas.isna(d3.data_ch).sum().sum() == 0

	d4 = DataFrames(ca=df, crack=True, ch=df.chose, wt='wgt')
	assert d4.data_wt is not None
	assert d4.data_wt.columns == 'wgt'
	assert d4.data_ch is not None
	assert d4.data_ch.shape == (5029, 6)
	assert d4.data_av is not None
	assert d4.data_av.shape == (5029, 6)

	d5 = DataFrames(ca=df, crack=True, ch=df.chose, wt=df.wgt)
	assert d5.data_wt is not None
	assert d5.data_wt.columns == 'wgt'
	assert d5.data_ch is not None
	assert d5.data_ch.shape == (5029, 6)
	assert d5.data_av is not None
	assert d5.data_av.shape == (5029, 6)
	assert d5.data_co.shape == (5029, 31)
	assert d5.data_ca is None
	assert d5.data_ce is not None
	assert d5.data_ce.shape == (22033, 5)

	with raises(ValueError):
		bad = DataFrames(co=df)

def test_dfs_init_co():
	from larch.data_warehouse import example_file
	raw_data = pandas.read_csv(example_file('swissmetro.csv.gz'))
	keep = raw_data.eval("PURPOSE in (1,3) and CHOICE != 0")
	selected_data = raw_data[keep]

	d0 = DataFrames(selected_data, alt_codes=[1, 2, 3])
	assert d0.data_co.shape == (6768, 28)
	assert d0.data_ca is None
	assert d0.data_ce is None
	assert d0.data_ch is None
	assert d0.data_av is None

	d1 = DataFrames(co=selected_data, alt_codes=[1, 2, 3])
	assert d1.data_co.shape == (6768, 28)
	assert d1.data_ca is None
	assert d1.data_ce is None
	assert d1.data_ch is None
	assert d1.data_av is None

	with raises(ValueError):
		DataFrames(ca=selected_data, alt_codes=[1, 2, 3])

	d2 = DataFrames(co=selected_data, alt_codes=[1, 2, 3], ch='CHOICE')
	assert d2.data_co.shape == (6768, 28)
	assert d2.data_ca is None
	assert d2.data_ce is None
	assert d2.data_ch is not None
	assert d2.data_ch.shape == (6768, 3)
	assert all(d2.data_ch.sum() == [908, 4090, 1770])
	assert d2.data_av is None

	d2 = DataFrames(co=selected_data, alt_codes=[1, 2, 3], ch=selected_data.CHOICE)
	assert d2.data_co.shape == (6768, 28)
	assert d2.data_ca is None
	assert d2.data_ce is None
	assert d2.data_ch is not None
	assert d2.data_ch.shape == (6768, 3)
	assert all(d2.data_ch.sum() == [908, 4090, 1770])
	assert d2.data_av is None

	d2 = DataFrames(co=selected_data, alt_codes=[1, 2, 3], ch='CHOICE', wt='GROUP')
	assert d2.data_co.shape == (6768, 28)
	assert d2.data_ca is None
	assert d2.data_ce is None
	assert d2.data_ch is not None
	assert d2.data_ch.shape == (6768, 3)
	assert all(d2.data_ch.sum() == [908, 4090, 1770])
	assert d2.data_av is None
	assert d2.data_wt is not None
	assert d2.data_wt.shape == (6768, 1)

	d2 = DataFrames(co=selected_data, alt_codes=[1, 2, 3], ch='CHOICE', wt=selected_data.GROUP)
	assert d2.data_co.shape == (6768, 28)
	assert d2.data_ca is None
	assert d2.data_ce is None
	assert d2.data_ch is not None
	assert d2.data_ch.shape == (6768, 3)
	assert all(d2.data_ch.sum() == [908, 4090, 1770])
	assert d2.data_av is None
	assert d2.data_wt is not None
	assert d2.data_wt.shape == (6768, 1)